import matplotlib.pyplot as plt

from raysect.optical.spectrum import Spectrum

from cherab.core.model.laser import SeldenMatobaThomsonSpectrum
from cherab.core.utility import RecursiveDict


density = 3e19  # electron density
temperatures = [100, 2.e3]  # electron temperatures in eV
laser_wavelength = 1060  # wavelength of the laser light in nm
laser_energy = 1  # energy density of the laser light in J/m^3

# angle between observation direction and electric field
angles_pol = [90, 45]

# angles between propagation direction and observation direction
angles_obs = [45, 90, 135, 120] 

# define spectrum of observed scattering
min_wavelength = 650
max_wavelength = 1300
bins = 1000


# calculate Thomson scattered spectra for the specified combinations of el. properties and observation
model = SeldenMatobaThomsonSpectrum()
scattered = RecursiveDict()

for te in temperatures:
    for ap in angles_pol:
        for ao in angles_obs:
            spectrum = Spectrum(min_wavelength, max_wavelength, bins)
            scattered[te][ap][ao] = model.calculate_spectrum(density, te, laser_energy, laser_wavelength,
                                                             ao, ap, spectrum).samples
scattered = scattered.freeze()

wvls = spectrum.wavelengths

# plot temperature influence on scattered spectra
ap, ao = 90, 90
_, ax = plt.subplots()
ax.set_title("Scattered spectrum, pol. angl.={:d}, obs. angl = {:d}".format(ap, ao))
for te in temperatures:
    rad = scattered[te][ap][ao]
    ax.plot(wvls, rad, label="Te = {:3.1f} eV".format(te))
ax.set_xlabel("Wavelength [nm]")
ax.set_ylabel("Spectral Radiance [W/m^3/sr]")
ax.legend()

# plot influence of angle between polarisation and observation on scattered spectra
te, ao = 100, 90
_, ax = plt.subplots()
ax.set_title("Scattered spectrum, te = {:3.1f}, obs. angl = {:d}".format(te, ao))
for ap in angles_pol:
    rad = scattered[te][ap][ao]
    ax.plot(wvls, rad, label="pol. angl. = {:d} deg".format(ap))
ax.set_xlabel("Wavelength [nm]")
ax.set_ylabel("Spectral Radiance [W/m^3/sr]")
ax.legend()

# plot influence of observation angle on scattered spectra
te, ap = 2e3, 90
_, ax = plt.subplots()
ax.set_title("Scattered spectrum, te = {:3.1f}, obs. angl = {:d}".format(te, ao))
for ao in angles_obs:
    rad = scattered[te][ap][ao]
    ax.plot(wvls, rad, label="obs. angl. = {:d} deg".format(ao))
ax.set_xlabel("Wavelength [nm]")
ax.set_ylabel("Spectral Radiance [W/m^3/sr]")
ax.legend()
plt.show()



# Copyright 2016-2018 Euratom
# Copyright 2016-2018 United Kingdom Atomic Energy Authority
# Copyright 2016-2018 Centro de Investigaciones Energéticas, Medioambientales y Tecnológicas
#
# Licensed under the EUPL, Version 1.1 or – as soon they will be approved by the
# European Commission - subsequent versions of the EUPL (the "Licence");
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at:
#
# https://joinup.ec.europa.eu/software/page/eupl5
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the Licence is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.
#
# See the Licence for the specific language governing permissions and limitations
# under the Licence.

# cython: language_level=3

from cherab.core.utility import Notifier

from libc.math cimport exp, M_PI
from raysect.optical cimport Vector3D
cimport cython

from cherab.core.math cimport autowrap_function3d, autowrap_vectorfunction3d
from cherab.core.utility.constants cimport ELEMENTARY_CHARGE


# must be immutable, once created cannot be modified as changes are not tracked.
cdef class DistributionFunction:
    """
    The distribution function base class.

    All plasma Species objects are defined through distribution functions. This base
    class defines the core interface which is used by the rest of the framework.
    This base class cannot be used on its own, it raises a NotImplementedError() on all
    method calls. Users should either use a simpler distribution instance
    (such as the Maxwellian distribution) or implement their own.
    """

    def __init__(self):
        self.notifier = Notifier()

    def __call__(self, double x, double y, double z, double vx, double vy, double vz):
        """
        Evaluates the phase space density at the specified point in 6D phase space.

        Wraps the cython evaluate() function for fast evaluation.

        :param float x: position in meters
        :param float y: position in meters
        :param float z: position in meters
        :param float vx: velocity in meters per second
        :param float vy: velocity in meters per second
        :param float vz: velocity in meters per second
        :return: phase space density in s^3/m^6
        """
        return self.evaluate(x, y, z, vx, vy, vz)

    cdef double evaluate(self, double x, double y, double z, double vx, double vy, double vz) except? -1e999:
        """
        Evaluates the phase space density at the specified point in 6D phase space.
        
        :param float x: position in meters
        :param float y: position in meters
        :param float z: position in meters
        :param float vx: velocity in meters per second
        :param float vy: velocity in meters per second
        :param float vz: velocity in meters per second
        :return: phase space density in s^3/m^6
        """
        raise NotImplementedError()

    cpdef Vector3D bulk_velocity(self, double x, double y, double z):
        """
        Evaluates the species' bulk velocity at the specified 3D coordinate.

        :param float x: position in meters
        :param float y: position in meters
        :param float z: position in meters
        :return: velocity vector in m/s
        :rtype: Vector3D
        """
        raise NotImplementedError()

    cpdef double effective_temperature(self, double x, double y, double z) except? -1e999:
        """
        Evaluates the species' effective temperature at the specified 3D coordinate.

        :param float x: position in meters
        :param float y: position in meters
        :param float z: position in meters
        :return: temperature in eV
        :rtype: float
        """
        raise NotImplementedError()

    cpdef double density(self, double x, double y, double z) except? -1e999:
        """
        Evaluates the species' density at the specified 3D coordinate.

        :param float x: position in meters
        :param float y: position in meters
        :param float z: position in meters
        :return: density in m^-3
        :rtype: float
        """
        raise NotImplementedError()


cdef class ZeroDistribution(DistributionFunction):
    """
    A zero distribution function.

    All distribution properties are zero.
    """

    def __init__(self):
        super().__init__()

    def __call__(self, double x, double y, double z, double vx, double vy, double vz):
        """
        Evaluates the phase space density at the specified point in 6D phase space.

        Wraps the cython evaluate() function for fast evaluation.

        :param float x: position in meters
        :param float y: position in meters
        :param float z: position in meters
        :param float vx: velocity in meters per second
        :param float vy: velocity in meters per second
        :param float vz: velocity in meters per second
        :return: phase space density 0 s^3/m^6
        """
        return self.evaluate(x, y, z, vx, vy, vz)

    cdef double evaluate(self, double x, double y, double z, double vx, double vy, double vz) except? -1e999:
        """
        Evaluates the phase space density at the specified point in 6D phase space.
        
        :param float x: position in meters
        :param float y: position in meters
        :param float z: position in meters
        :param float vx: velocity in meters per second
        :param float vy: velocity in meters per second
        :param float vz: velocity in meters per second
        :return: phase space density 0 s^3/m^6
        """
        return 0.0

    cpdef Vector3D bulk_velocity(self, double x, double y, double z):
        """
        Evaluates the species' bulk velocity at the specified 3D coordinate.

        :param float x: position in meters
        :param float y: position in meters
        :param float z: position in meters
        :return: velocity vector (0, 0, 0) in m/s
        :rtype: Vector3D
        """
        return Vector3D(0, 0, 0)

    cpdef double effective_temperature(self, double x, double y, double z) except? -1e999:
        """
        Returns 0 species' effective temperature at the specified 3D coordinate.

        :param float x: position in meters
        :param float y: position in meters
        :param float z: position in meters
        :return: temperature 0 eV
        :rtype: float
        """
        return 0.0

    cpdef double density(self, double x, double y, double z) except? -1e999:
        """
        Returns 0 species' density.

        :param float x: position in meters
        :param float y: position in meters
        :param float z: position in meters
        :return: density 0 m^-3
        :rtype: float
        """
        return 0.0


cdef class Maxwellian(DistributionFunction):
    """
    A Maxwellian distribution function.

    This class implements a Maxwell-Boltzmann distribution, the statistical distribution
    describing a system of particles that have reached thermodynamic equilibrium. The
    user supplies 3D functions that provide the mean density, temperature and velocity
    respectively.

    :param Function3D density: 3D function defining the density in cubic meters.
    :param Function3D temperature: 3D function defining the temperature in eV.
    :param VectorFunction3D velocity: 3D vector function defining the bulk velocity in meters per second.
    :param double atomic_mass: Atomic mass of the species in kg.

    .. code-block:: pycon

       >>> from scipy.constants import atomic_mass
       >>> from raysect.core.math import Vector3D
       >>> from cherab.core import Maxwellian
       >>> from cherab.core.atomic import deuterium
       >>>
       >>> # Setup distribution for a slab of plasma in thermodynamic equilibrium
       >>> d0_density = 1E17
       >>> d0_temperature = 1
       >>> bulk_velocity = Vector3D(0, 0, 0)
       >>> d0_distribution = Maxwellian(d0_density, d0_temperature, bulk_velocity, deuterium.atomic_weight * atomic_mass)
    """

    def __init__(self, object density, object temperature, object velocity, double atomic_mass):

        super().__init__()
        self._density = autowrap_function3d(density)
        self._temperature = autowrap_function3d(temperature)
        self._velocity = autowrap_vectorfunction3d(velocity)
        self._atomic_mass = atomic_mass

    @cython.cdivision(True)
    cdef double evaluate(self, double x, double y, double z, double vx, double vy, double vz) except? -1e999:
        """
        Evaluates the phase space density at the specified point in 6D phase space.

        :param x: position in meters
        :param y: position in meters
        :param z: position in meters
        :param vx: velocity in meters per second
        :param vy: velocity in meters per second
        :param vz: velocity in meters per second
        :return: phase space density in s^3/m^6
        """

        cdef:
            double k1, k2, ux, uy, uz
            Vector3D bulk_velocity

        k1 = self._atomic_mass / (2 * ELEMENTARY_CHARGE * self._temperature.evaluate(x, y, z))
        k2 = (k1 / M_PI) ** 1.5

        bulk_velocity = self._velocity.evaluate(x, y, z)
        ux = vx - bulk_velocity.x
        uy = vy - bulk_velocity.y
        uz = vz - bulk_velocity.z

        return self._density.evaluate(x, y, z) * exp(-k1 * (ux*ux + uy*uy + uz*uz)) * k2

    cpdef Vector3D bulk_velocity(self, double x, double y, double z):
        """
        Evaluates the species' bulk velocity at the specified 3D coordinate.

        :param x: position in meters
        :param y: position in meters
        :param z: position in meters
        :return: velocity vector in m/s
        
        .. code-block:: pycon

           >>> d0_distribution.bulk_velocity(1, 0, 0)
           Vector3D(0.0, 0.0, 0.0)
        """

        return self._velocity.evaluate(x, y, z)

    cpdef double effective_temperature(self, double x, double y, double z) except? -1e999:
        """

        :param x: position in meters
        :param y: position in meters
        :param z: position in meters
        :return: temperature in eV
        
        .. code-block:: pycon
        
           >>> d0_distribution.effective_temperature(1, 0, 0)
           1.0
        """

        return self._temperature.evaluate(x, y, z)

    cpdef double density(self, double x, double y, double z) except? -1e999:
        """

        :param x: position in meters
        :param y: position in meters
        :param z: position in meters
        :return: density in m^-3

        .. code-block:: pycon

           >>> d0_distribution.density(1, 0, 0)
           1e+17
        """

        return self._density.evaluate(x, y, z)





# Copyright 2016-2018 Euratom
# Copyright 2016-2018 United Kingdom Atomic Energy Authority
# Copyright 2016-2018 Centro de Investigaciones Energéticas, Medioambientales y Tecnológicas
#
# Licensed under the EUPL, Version 1.1 or – as soon they will be approved by the
# European Commission - subsequent versions of the EUPL (the "Licence");
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at:
#
# https://joinup.ec.europa.eu/software/page/eupl5
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the Licence is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.
#
# See the Licence for the specific language governing permissions and limitations
# under the Licence.

from cherab.core.math.function cimport Function2D
from numpy cimport ndarray, int8_t


cdef class Caching2D(Function2D):

    cdef readonly:
        Function2D function
        int no_boundary_error
        ndarray x_np, y_np
        double[::1] x_domain_view, y_domain_view
        int top_index_x, top_index_y
        double x_min, x_delta_inv, y_min, y_delta_inv
        double data_min, data_max, data_delta, data_delta_inv
        double[::1] x_view, x2_view, x3_view
        double[::1] y_view, y2_view, y3_view
        double[:,::1] data_view
        double[:,:,::1] coeffs_view
        int8_t[:,::1] calculated_view

    cdef double evaluate(self, double px, double py) except? -1e999

    cdef double _evaluate(self, double px, double py, int i_x, int i_y)

    cdef double _evaluate_polynomial_derivative(self, int i_x, int i_y, double px, double py, int der_x, int der_y)



# Copyright 2016-2023 Euratom
# Copyright 2016-2023 United Kingdom Atomic Energy Authority
# Copyright 2016-2023 Centro de Investigaciones Energéticas, Medioambientales y Tecnológicas
#
# Licensed under the EUPL, Version 1.1 or – as soon they will be approved by the
# European Commission - subsequent versions of the EUPL (the "Licence");
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at:
#
# https://joinup.ec.europa.eu/software/page/eupl5
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the Licence is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.
#
# See the Licence for the specific language governing permissions and limitations
# under the Licence.

import unittest

import numpy as np

from raysect.core import Point3D, Vector3D
from raysect.optical import Ray, World

from cherab.core.atomic import AtomicData, LineRadiationPower, ContinuumPower, CXRadiationPower
from cherab.core.atomic import deuterium, hydrogen, nitrogen
from cherab.tools.plasmas.slab import build_constant_slab_plasma
from cherab.core.model import TotalRadiatedPower

from cherab.core.utility import EvAmuToMS, EvToJ


class ConstantLineRadiationPower(LineRadiationPower):
    """
    Constant line radiation power coefficient.
    """

    def __init__(self, value):
        self.value = value

    def evaluate(self, density, temperature):

        return self.value


class ConstantContinuumPower(ContinuumPower):
    """
    Constant continuum power coefficient.
    """

    def __init__(self, value):
        self.value = value

    def evaluate(self, density, temperature):

        return self.value


class ConstantCXRadiationPower(CXRadiationPower):
    """
    Constant charge exchange radiation power coefficient.
    """

    def __init__(self, value):
        self.value = value

    def evaluate(self, density, temperature):

        return self.value


class MockAtomicData(AtomicData):
    """Fake atomic data for test purpose."""

    def line_radiated_power_rate(self, element, charge):

        return ConstantLineRadiationPower(1.e-32)

    def continuum_radiated_power_rate(self, element, charge):

        return ConstantContinuumPower(1.e-33)

    def cx_radiated_power_rate(self, element, charge):

        return ConstantCXRadiationPower(1.e-31)


class TestTotalRadiatedPower(unittest.TestCase):

    def setUp(self):

        self.world = World()

        self.atomic_data = MockAtomicData()

        plasma_species = [(deuterium, 0, 1.e18, 500., Vector3D(0, 0, 0)),
                          (hydrogen, 0, 1.e18, 500., Vector3D(0, 0, 0)),
                          (nitrogen, 6, 5.e18, 1100., Vector3D(0, 0, 0)),
                          (nitrogen, 7, 1.e19, 1100., Vector3D(0, 0, 0))]
        self.plasma = build_constant_slab_plasma(length=1.2, width=1.2, height=1.2,
                                                 electron_density=1e19, electron_temperature=1000.,
                                                 plasma_species=plasma_species)
        self.plasma.parent = self.world
        self.plasma.atomic_data = self.atomic_data

    def test_total_radiated_power(self):

        self.plasma.models = [TotalRadiatedPower(nitrogen, 6)]

        # observing
        origin = Point3D(1.5, 0, 0)
        direction = Vector3D(-1, 0, 0)
        ray = Ray(origin=origin, direction=direction,
                  min_wavelength=500., max_wavelength=550., bins=2)
        radiated_power = ray.trace(self.world).total()

        # validating
        ne = self.plasma.electron_distribution.density(0.5, 0.5, 0.5)
        n_n6 = self.plasma.composition[(nitrogen, 6)].distribution.density(0.5, 0.5, 0.5)
        n_n7 = self.plasma.composition[(nitrogen, 7)].distribution.density(0.5, 0.5, 0.5)
        n_h0 = self.plasma.composition[(hydrogen, 0)].distribution.density(0.5, 0.5, 0.5)
        n_d0 = self.plasma.composition[(deuterium, 0)].distribution.density(0.5, 0.5, 0.5)

        integration_length = 1.2

        plt_rate = self.atomic_data.line_radiated_power_rate(nitrogen, 6).value
        plt_radiance = 0.25 / np.pi * plt_rate * ne * n_n6 * integration_length

        prb_rate = self.atomic_data.continuum_radiated_power_rate(nitrogen, 7).value
        prb_radiance = 0.25 / np.pi * prb_rate * ne * n_n7 * integration_length

        prc_rate = self.atomic_data.cx_radiated_power_rate(nitrogen, 7).value
        prc_radiance = 0.25 / np.pi * prc_rate * (n_h0 + n_d0) * n_n7 * integration_length

        test_radiated_power = plt_radiance + prb_radiance + prc_radiance

        self.assertAlmostEqual(radiated_power / test_radiated_power, 1., delta=1e-8)


if __name__ == '__main__':
    unittest.main()



# cython: language_level=3

# Copyright 2016-2023 Euratom
# Copyright 2016-2023 United Kingdom Atomic Energy Authority
# Copyright 2016-2023 Centro de Investigaciones Energéticas, Medioambientales y Tecnológicas
#
# Licensed under the EUPL, Version 1.1 or – as soon they will be approved by the
# European Commission - subsequent versions of the EUPL (the "Licence");
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at:
#
# https://joinup.ec.europa.eu/software/page/eupl5
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the Licence is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.
#
# See the Licence for the specific language governing permissions and limitations
# under the Licence.

import numpy as np

from raysect.optical cimport Spectrum, Point3D, Vector3D

from cherab.core.atomic cimport Line, AtomicData
from cherab.core.species cimport Species
from cherab.core.plasma cimport Plasma
from cherab.core.model.lineshape.doppler cimport doppler_shift, thermal_broadening
from cherab.core.model.lineshape.gaussian cimport add_gaussian_line

cimport cython

# required by numpy c-api
np.import_array()


DEF MULTIPLET_WAVELENGTH = 0
DEF MULTIPLET_RATIO = 1


cdef class MultipletLineShape(LineShapeModel):
    """
    Produces Multiplet line shapes.

    The lineshape radiance is calculated from a base PEC rate that is unresolved. This
    radiance is then divided over a number of components as specified in the multiplet
    argument. The multiplet components are specified with an Nx2 array where N is the
    number of components in the multiplet. The first axis of the array contains the
    wavelengths of each component, the second contains the line ratio for each component.
    The component line ratios must sum to one. For example:

    :param Line line: The emission line object for the base rate radiance calculation.
    :param float wavelength: The rest wavelength of the base emission line.
    :param Species target_species: The target plasma species that is emitting.
    :param Plasma plasma: The emitting plasma object.
    :param AtomicData atomic_data: The atomic data provider.
    :param multiplet: An Nx2 array that specifies the multiplet wavelengths and line ratios.

    .. code-block:: pycon

       >>> from cherab.core.atomic import Line, nitrogen
       >>> from cherab.core.model import ExcitationLine, MultipletLineShape
       >>>
       >>> # multiplet specification in Nx2 array
       >>> multiplet = [[403.509, 404.132, 404.354, 404.479, 405.692], [0.205, 0.562, 0.175, 0.029, 0.029]]
       >>>
       >>> # Adding the multiplet to the plasma model.
       >>> nitrogen_II_404 = Line(nitrogen, 1, ("2s2 2p1 4f1 3G13.0", "2s2 2p1 3d1 3F10.0"))
       >>> excit = ExcitationLine(nitrogen_II_404, lineshape=MultipletLineShape, lineshape_args=[multiplet])
       >>> plasma.models.add(excit)
    """

    def __init__(self, Line line, double wavelength, Species target_species, Plasma plasma, AtomicData atomic_data,
                 object multiplet):

        super().__init__(line, wavelength, target_species, plasma, atomic_data)

        multiplet = np.array(multiplet, dtype=np.float64)

        if not (len(multiplet.shape) == 2 and multiplet.shape[0] == 2):
            raise ValueError("The multiplet specification must be an array of shape (Nx2).")

        if not multiplet[1, :].sum() == 1.0:
            raise ValueError("The multiplet line ratios should sum to one.")

        self._number_of_lines = multiplet.shape[1]
        self._multiplet = multiplet
        self._multiplet_mv = self._multiplet

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.initializedcheck(False)
    cpdef Spectrum add_line(self, double radiance, Point3D point, Vector3D direction, Spectrum spectrum):

        cdef double ts, sigma, shifted_wavelength, component_wavelength, component_radiance
        cdef Vector3D ion_velocity

        ts = self.target_species.distribution.effective_temperature(point.x, point.y, point.z)
        if ts <= 0.0:
            return spectrum

        ion_velocity = self.target_species.distribution.bulk_velocity(point.x, point.y, point.z)

        # calculate the line width
        sigma = thermal_broadening(self.wavelength, ts, self.line.element.atomic_weight)

        for i in range(self._number_of_lines):

            component_wavelength = self._multiplet_mv[MULTIPLET_WAVELENGTH, i]
            component_radiance = radiance * self._multiplet_mv[MULTIPLET_RATIO, i]

            # calculate emission line central wavelength, doppler shifted along observation direction
            shifted_wavelength = doppler_shift(component_wavelength, direction, ion_velocity)

            spectrum = add_gaussian_line(component_radiance, shifted_wavelength, sigma, spectrum)

        return spectrum



"""
This example demonstrates using a grid of voxels to invert bolometer
measurements and produce a 2D emissivity profile. The bolometer
measurements are provided by observations of a radiating plasma, with
the radiation defined by a 2D function. The inverted emissivity profile
is compared with the original phantom to show the quality of the
inversion.

The bolometer and radiating emitter are produced in the same way as the
`observe_radiation_function.py` demo. We use a simplified cylindrical
geometry and the same radiation function as in the radiation_function.py
demo.
"""
import math
from pathlib import Path
import pickle
import matplotlib.pyplot as plt
import numpy as np

from raysect.core import Node, Point2D, Point3D, Vector3D, rotate_basis, rotate_y, translate
from raysect.optical import World
from raysect.optical.material import AbsorbingSurface, VolumeTransform
from raysect.primitive import Box, Cylinder, Subtract

from cherab.core.math import AxisymmetricMapper
from cherab.tools.emitters import RadiationFunction
from cherab.tools.inversions import ToroidalVoxelGrid, invert_regularised_nnls
from cherab.tools.observers import BolometerCamera, BolometerSlit, BolometerFoil


# Convenient constants
XAXIS = Vector3D(1, 0, 0)
YAXIS = Vector3D(0, 1, 0)
ZAXIS = Vector3D(0, 0, 1)
ORIGIN = Point3D(0, 0, 0)

# Bolometer geometry: same as the demo which generates the geometry matrix
BOX_WIDTH = 0.05
BOX_WIDTH = 0.2
BOX_HEIGHT = 0.07
BOX_DEPTH = 0.2
SLIT_WIDTH = 0.004
SLIT_HEIGHT = 0.005
FOIL_WIDTH = 0.0013
FOIL_HEIGHT = 0.0038
FOIL_CORNER_CURVATURE = 0.0005
SLIT_SENSOR_SEPARATION = 0.1
FOIL_SEPARATION = 0.00508  # 0.2 inch between foils
SENSOR_ANGLES = [-18, -6, 6, 18]

world = World()


################################################################################
# Make a simple vessel geometry
################################################################################
centre_column_radius = 1
vessel_wall_radius = 3.7
vessel_height = 3.7
vessel = Subtract(
    Cylinder(radius=vessel_wall_radius, height=vessel_height),
    Cylinder(radius=centre_column_radius, height=vessel_height),
    material=AbsorbingSurface(), name="Vessel", parent=world,
)


################################################################################
# Build a simple bolometer system
################################################################################
def make_bolometer_camera():
    # The camera consists of a box with a rectangular slit and 4 sensors, each
    # of which has 4 foils.
    # In its local coordinate system, the camera's slit is located at the
    # origin and the sensors below the z=0 plane, looking up towards the slit.
    camera_box = Box(lower=Point3D(-BOX_WIDTH / 2, -BOX_HEIGHT / 2, -BOX_DEPTH),
                     upper=Point3D(BOX_WIDTH / 2, BOX_HEIGHT / 2, 0))
    # Hollow out the box
    inside_box = Box(lower=camera_box.lower + Vector3D(1e-5, 1e-5, 1e-5),
                     upper=camera_box.upper - Vector3D(1e-5, 1e-5, 1e-5))
    camera_box = Subtract(camera_box, inside_box)
    # The slit is a hole in the box
    aperture = Box(lower=Point3D(-SLIT_WIDTH / 2, -SLIT_HEIGHT / 2, -1e-4),
                   upper=Point3D(SLIT_WIDTH / 2, SLIT_HEIGHT / 2, 1e-4))
    camera_box = Subtract(camera_box, aperture)
    camera_box.material = AbsorbingSurface()
    # Instance of the bolometer camera
    bolometer_camera = BolometerCamera(camera_geometry=camera_box)
    # The bolometer slit in this instance just contains targeting information
    # for the ray tracing, since we have already given our camera a geometry
    # The slit is defined in the local coordinate system of the camera
    slit = BolometerSlit(slit_id="Example slit", centre_point=ORIGIN,
                         basis_x=XAXIS, dx=SLIT_WIDTH, basis_y=YAXIS, dy=SLIT_HEIGHT,
                         parent=bolometer_camera, csg_aperture=False)
    for j, angle in enumerate(SENSOR_ANGLES):
        # 4 bolometer foils, spaced at equal intervals along the local X axis
        # The bolometer positions and orientations are given in the local coordinate
        # system of the camera, just like the slit
        sensor = Node(name="Bolometer sensor", parent=bolometer_camera,
                      transform=rotate_y(angle) * translate(0, 0, -SLIT_SENSOR_SEPARATION))
        # The foils are shifted relative to the centre of the sensor by -1.5, -0.5, 0.5 and 1.5
        # times the foil-foil separation
        for i, shift in enumerate([-1.5, -0.5, 0.5, 1.5]):
            foil_transform = sensor.transform * translate(shift * FOIL_SEPARATION, 0, 0)
            foil = BolometerFoil(detector_id="Foil {} sensor {}".format(i + 1, j + 1),
                                 centre_point=ORIGIN.transform(foil_transform),
                                 basis_x=XAXIS.transform(foil_transform), dx=FOIL_WIDTH,
                                 basis_y=YAXIS.transform(foil_transform), dy=FOIL_HEIGHT,
                                 slit=slit, parent=bolometer_camera, units="Power",
                                 accumulate=False, curvature_radius=FOIL_CORNER_CURVATURE)
            bolometer_camera.add_foil_detector(foil)
    return bolometer_camera


# Make several cameras distributed around the outside of the vessel
camera_angles = [20, 60, 100]
rotation_origin = Point2D(1.5, 1.5)
cameras = []
for camera_angle in camera_angles:
    camera = make_bolometer_camera()
    camera.transform = (translate(rotation_origin.x, 0, rotation_origin.y)
                        * rotate_y(camera_angle)
                        * translate(0, 0, 2.0)
                        * rotate_basis(-ZAXIS, YAXIS)
                        )
    camera.parent = world
    camera.name = "Angle {}".format(camera_angle)
    cameras.append(camera)


################################################################################
# Produce a simple radiating plasma.
################################################################################

# This is the same emission profile used in the radiation_function.py demo,
# but scaled down by 1e6 to make the units on the plots cleaner
PLASMA_AXIS = Point2D(1.5, 1.5)
LCFS_RADIUS = 1
RING_RADIUS = 0.5

RADIATION_PEAK = 1
CENTRE_PEAK_WIDTH = 0.05
RING_WIDTH = 0.025

# distance of wall from LCFS
WALL_LCFS_OFFSET = 0.1

CYLINDER_RADIUS = PLASMA_AXIS.x + LCFS_RADIUS + WALL_LCFS_OFFSET * 1.1
CYLINDER_HEIGHT = (LCFS_RADIUS + WALL_LCFS_OFFSET) * 2


def emission_function_2d(r, z):

    sample_point = Point2D(r, z)
    direction = PLASMA_AXIS.vector_to(sample_point)
    bearing = math.atan2(direction.y, direction.x)

    # calculate radius of coordinate from magnetic axis
    radius_from_axis = direction.length
    closest_ring_point = PLASMA_AXIS + (direction.normalise() * 0.5)
    radius_from_ring = sample_point.distance_to(closest_ring_point)

    # evaluate pedestal-> core function
    if radius_from_axis <= LCFS_RADIUS:

        central_radiatior = RADIATION_PEAK * math.exp(-(radius_from_axis ** 2) / CENTRE_PEAK_WIDTH)

        ring_radiator = (RADIATION_PEAK * math.cos(bearing)
                         * math.exp(-(radius_from_ring ** 2) / RING_WIDTH))
        ring_radiator = max(0, ring_radiator)

        return central_radiatior + ring_radiator
    else:
        return 0


emitter = Cylinder(radius=CYLINDER_RADIUS, height=CYLINDER_HEIGHT,
                   transform=translate(0, 0, 0))
emission_function_3d = AxisymmetricMapper(emission_function_2d)
emitting_material = VolumeTransform(RadiationFunction(emission_function_3d),
                                    transform=emitter.transform.inverse())
emitter.material = emitting_material
emitter.parent = world

################################################################################
# Measure the radiation with the bolometers
################################################################################
print("Measuring the radiation with the bolometers...")
observations = []
for camera in cameras:
    observations.extend(camera.observe())
observations = np.asarray(observations)


################################################################################
# Load the geometry data required for the inversion
################################################################################
script_dir = Path(__file__).parent
try:
    with open(script_dir / "voxel_grid_data.pickle", "rb") as f:
        grid_data = pickle.load(f)
except FileNotFoundError:
    raise RuntimeError(
        "Geometry data not found: please run the "
        "geometry_matrix_with_voxels.py demo first."
    )

voxel_grid = ToroidalVoxelGrid(grid_data['voxel_data'])
grid_laplacian = grid_data['laplacian']
sensitivity_matrix = grid_data['sensitivity_matrix']

# The sensitivity matrix has units of [m3sr] when the units used to calculate
# the sensitivity in BolometerFoil.calculate_sensitivity are "Power".  The
# radiation function above defines an isotropic emissivity with units of [Wm-3],
# so we need to divide the sensitivity matrix by 4pi sr to use with power
# measurements in [W].
sensitivity_matrix /= 4 * math.pi

################################################################################
# Perform the inversion
################################################################################
print("Performing inversion...")
# We'll use a regularised NNLS inversion here: experience shows it performs
# better than SART for bolometry systems with relatively low numbers of
# foils (though it does perform well for imaging cameras with high resolution).

# Apply a large penalty for placing radiation where there is no bolometer
# coverage, else the algorithm can put radiation there without affecting the
# measurements
grid_laplacian[:, np.nonzero(sensitivity_matrix.sum(axis=0) == 0)] = 1e10

inverted_emission, residual_norm = invert_regularised_nnls(
    sensitivity_matrix, observations, alpha=1e-11, tikhonov_matrix=grid_laplacian
)

################################################################################
# Plot the emissivity profile and the inverted profile
################################################################################
print("Plotting results...")
phantom_samples = voxel_grid.emissivities_from_function(emission_function_3d)
fig, (ax1, ax2) = plt.subplots(1, 2)
voxel_grid.plot(ax=ax1, voxel_values=phantom_samples, title="Phantom")
voxel_grid.plot(ax=ax2, voxel_values=inverted_emission, title="Inversion")
ax1.set_xlabel("r")
ax1.set_ylabel("z")
ax2.set_xlabel("r")
ax2.set_ylabel("z")
fig.tight_layout()

################################################################################
# Calculate some metrics for comparison with the phantom
################################################################################
back_calculated_measurements = sensitivity_matrix @ inverted_emission
phantom_measurements = sensitivity_matrix @ phantom_samples
cell_volumes = np.fromiter((cell.volume for cell in voxel_grid), 'float')
total_phantom_power = phantom_samples @ cell_volumes
total_inversion_power = inverted_emission @ cell_volumes

fig, ax = plt.subplots()
ax.plot(observations, linestyle="-", label="Ray-traced measurements")
ax.plot(phantom_measurements, linestyle="--", label="Voxel-based measurements")
ax.plot(back_calculated_measurements, linestyle="--", label="Back-calculated from inversion")
ax.set_xlabel("Foil")
ax.set_ylabel("Power / W")
ax.legend()

print("Phantom total power: {:.4g}W".format(total_phantom_power))
print("Inversion total power: {:.4g}W".format(total_inversion_power))

plt.show()



# cython: language_level=3

# Copyright 2016-2023 Euratom
# Copyright 2016-2023 United Kingdom Atomic Energy Authority
# Copyright 2016-2023 Centro de Investigaciones Energéticas, Medioambientales y Tecnológicas
#
# Licensed under the EUPL, Version 1.1 or – as soon they will be approved by the
# European Commission - subsequent versions of the EUPL (the "Licence");
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at:
#
# https://joinup.ec.europa.eu/software/page/eupl5
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the Licence is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.
#
# See the Licence for the specific language governing permissions and limitations
# under the Licence.

from scipy.special import hyp2f1

from libc.math cimport sqrt, floor, ceil, fabs, log, exp
from raysect.core.math.function.float cimport Function1D
from raysect.optical cimport Point3D, Vector3D

from cherab.core.atomic cimport Line, AtomicData
from cherab.core.species cimport Species
from cherab.core.plasma cimport Plasma
from cherab.core.atomic.elements import hydrogen, deuterium, tritium
from cherab.core.math.function cimport autowrap_function1d, autowrap_function2d
from cherab.core.math.integrators cimport GaussianQuadrature
from cherab.core.utility.constants cimport BOHR_MAGNETON, HC_EV_NM
from cherab.core.model.lineshape.doppler cimport doppler_shift, thermal_broadening
from cherab.core.model.lineshape.gaussian cimport add_gaussian_line


cimport cython


DEF PI_POLARISATION = 0
DEF SIGMA_POLARISATION = 1
DEF SIGMA_PLUS_POLARISATION = 1
DEF SIGMA_MINUS_POLARISATION = -1
DEF NO_POLARISATION = 2

DEF LORENTZIAN_CUTOFF_GAMMA = 50.0

cdef double _SIGMA2FWHM = 2 * sqrt(2 * log(2))


cdef class StarkFunction(Function1D):
    """
    Normalised Stark function for the StarkBroadenedLine line shape.

    :param float wavelength: central wavelength of the line in nm.
    :param float lambda_1_2: FWHM of the function in nm.
    """

    cdef double _a, _x0, _norm

    STARK_NORM_COEFFICIENT = 4 * LORENTZIAN_CUTOFF_GAMMA * hyp2f1(0.4, 1, 1.4, -(2 * LORENTZIAN_CUTOFF_GAMMA)**2.5)

    def __init__(self, double wavelength, double lambda_1_2):

        if wavelength <= 0:
            raise ValueError("Argument 'wavelength' must be positive.")

        if lambda_1_2 <= 0:
            raise ValueError("Argument 'lambda_1_2' must be positive.")

        self._x0 = wavelength
        self._a = (0.5 * lambda_1_2)**2.5
        # normalise, so the integral over x is equal to 1 in the limits
        # (_x0 - LORENTZIAN_CUTOFF_GAMMA * lambda_1_2, _x0 + LORENTZIAN_CUTOFF_GAMMA * lambda_1_2)
        self._norm = (0.5 * lambda_1_2)**1.5 / <double> self.STARK_NORM_COEFFICIENT

    @cython.cdivision(True)
    cdef double evaluate(self, double x) except? -1e999:

        return self._norm / ((fabs(x - self._x0))**2.5 + self._a)


@cython.cdivision(True)
@cython.initializedcheck(False)
@cython.boundscheck(False)
@cython.wraparound(False)
cpdef Spectrum add_lorentzian_line(double radiance, double wavelength, double lambda_1_2, Spectrum spectrum, Integrator1D integrator):
    r"""
    Adds a modified Lorentzian line to the given spectrum and returns the new spectrum.

    The modified Lorentzian:
    :math:`L(\lambda-\lambda_0, \Delta\lambda_{1/2}^{(L)})=\frac{C_0(\Delta\lambda_{1/2}^{(L)})^{3/2}}{(\lambda-\lambda_0)^{5/2}+(\frac{\Delta\lambda_{1/2}^{(L)}}{2})^{5/2}},`

    :math:`C_0=\frac{(1/2)^{3/2}}{4R{_2}F_1(\frac{2}{5},1,\frac{7}{5},-R^{5/2})},`

    where :math:`\Delta\lambda_{1/2}^{(L)}` is the line FWHM, :math:`{_2}F_1` is the hypergeometric function and :math:`R=50`.
    The line shape is truncated at :math:`\lambda<R\Delta\lambda_{1/2}^{(L)}` and :math:`\lambda>R\Delta\lambda_{1/2}^{(L)}`.

    See B. Lomanowski, et al. "Inferring divertor plasma properties from hydrogen Balmer
    and Paschen series spectroscopy in JET-ILW." Nuclear Fusion 55.12 (2015)
    `123028 <https://doi.org/10.1088/0029-5515/55/12/123028>`_ for details.

    :param float radiance: Intensity of the line in radiance.
    :param float wavelength: central wavelength of the line in nm.
    :param float lambda_1_2: FWHM of the line shape in nm.
    :param Spectrum spectrum: the current spectrum to which the Lorentzian line is added.
    :param Integrator1D integrator: Integrator1D instance to integrate the line shape
                                    over the spectral bin.
    :return:
    """

    cdef double cutoff_lower_wavelength, cutoff_upper_wavelength
    cdef double lower_wavelength, upper_wavelength
    cdef double bin_integral
    cdef int start, end, i

    if lambda_1_2 <= 0:
        return spectrum

    integrator.function = StarkFunction(wavelength, lambda_1_2)

    # calculate and check end of limits
    cutoff_lower_wavelength = wavelength - LORENTZIAN_CUTOFF_GAMMA * lambda_1_2
    if spectrum.max_wavelength < cutoff_lower_wavelength:
        return spectrum

    cutoff_upper_wavelength = wavelength + LORENTZIAN_CUTOFF_GAMMA * lambda_1_2
    if spectrum.min_wavelength > cutoff_upper_wavelength:
        return spectrum

    # locate range of bins where there is significant contribution from the gaussian (plus a health margin)
    start = max(0, <int> floor((cutoff_lower_wavelength - spectrum.min_wavelength) / spectrum.delta_wavelength))
    end = min(spectrum.bins, <int> ceil((cutoff_upper_wavelength - spectrum.min_wavelength) / spectrum.delta_wavelength))

    # add line to spectrum
    lower_wavelength = spectrum.min_wavelength + start * spectrum.delta_wavelength

    for i in range(start, end):
        upper_wavelength = spectrum.min_wavelength + spectrum.delta_wavelength * (i + 1)

        bin_integral = integrator.evaluate(lower_wavelength, upper_wavelength)
        spectrum.samples_mv[i] += radiance * bin_integral / spectrum.delta_wavelength

        lower_wavelength = upper_wavelength

    return spectrum


cdef class StarkBroadenedLine(ZeemanLineShapeModel):
    r"""
    Parametrised Stark-Doppler-Zeeman line shape for Balmer and Paschen series based on
    B. Lomanowski, et al. "Inferring divertor plasma properties from hydrogen Balmer
    and Paschen series spectroscopy in JET-ILW." Nuclear Fusion 55.12 (2015)
    `123028 <https://doi.org/10.1088/0029-5515/55/12/123028>`_.

    The following approximations are used:

      * The Zeeman and Stark effects are considered independently.
      * Zeeman splitting is taken in the form of a simple triplet with a :math:`\pi`-component
        centred at :math:`\lambda`, :math:`\sigma^{+}`-component at :math:`\frac{hc}{hc/\lambda -\mu B}`
        and :math:`\sigma^{-}`-component at :math:`\frac{hc}{hc/\lambda +\mu B}`.
      * The model of Stark broadening is obtained by fitting the Model Microfield Method (MMM).
      * The convolution of Stark-Zeeman and Doppler profiles is replaced with the weighted sum
        to speed-up calculations (so-called pseudo-Voigt profile).

    The Stark-broadened line shape is modelled as modified Lorentzian:
    :math:`L(\lambda-\lambda_0, \Delta\lambda_{1/2}^{(L)})=\frac{C_0(\Delta\lambda_{1/2}^{(L)})^{3/2}}{(\lambda-\lambda_0)^{5/2}+(\frac{\Delta\lambda_{1/2}^{(L)}}{2})^{5/2}},`

    :math:`C_0=\frac{(1/2)^{3/2}}{4R{_2}F_1(\frac{2}{5},1,\frac{7}{5},-R^{5/2})},`

    where :math:`\Delta\lambda_{1/2}^{(L)}=c_{ij}\frac{n_e^{a_{ij}}}{T_e^{b_{ij}}}` is the line FWHM,
    :math:`{_2}F_1` is the hypergeometric function and :math:`R=50`. 
    The line shape is truncated at :math:`\lambda<R\Delta\lambda_{1/2}^{(L)}` and :math:`\lambda>R\Delta\lambda_{1/2}^{(L)}`.
    The :math:`a_{ij}`, :math:`b_{ij}` and :math:`c_{ij}` are the fitting coefficients.

    Each Zeeman component is modelled as a weighted sum of Stark (Lorentzian), :math:`L(\lambda-\lambda_0, \Delta\lambda_{1/2}^{(L)})`,
    and Doppler (Gauss), :math:`G(\lambda'-\lambda_0, \Delta\lambda_{1/2}^{(G)})` profiles:

      :math:`\eta L(\lambda-\lambda_0, \Delta\lambda_{1/2}^{(V)}) + (1-\eta)G(\lambda-\lambda_0, \Delta\lambda_{1/2}^{(V)})`,
    
      with :math:`\Delta\lambda_{1/2}^{(V)}\equiv \Delta\lambda_{1/2}^{(V)}(\Delta\lambda_{1/2}^{(G)}, \Delta\lambda_{1/2}^{(L)})`
      and :math:`\eta\equiv \eta(\Delta\lambda_{1/2}^{(G)}, \Delta\lambda_{1/2}^{(L)})`.

    Both :math:`\Delta\lambda_{1/2}^{(V)}` and :math:`\eta` are obtained by fitting the convolution:

      :math:`\int_{-\infty}^{\infty}G(\lambda'-\lambda_0, \Delta\lambda_{1/2}^{(G)})L(\lambda-\lambda'-\lambda_0, \Delta\lambda_{1/2}^{(L)})d\lambda'`.

    The :math:`\Delta\lambda_{1/2}^{(V)}` function is fitted as:
    :math:`\Delta\lambda_{1/2}^{(V)}=\sum_{n=0}^{6}a_n(\frac{\Delta\lambda_{1/2}^{(L)}}{\Delta\lambda_{1/2}^{(G)}})^n`
    for :math:`\frac{\Delta\lambda_{1/2}^{(L)}}{\Delta\lambda_{1/2}^{(G)}} \le 1` and
    :math:`\Delta\lambda_{1/2}^{(V)}=\sum_{n=0}^{6}b_n(\frac{\Delta\lambda_{1/2}^{(G)}}{\Delta\lambda_{1/2}^{(L)}})^n`
    for :math:`\frac{\Delta\lambda_{1/2}^{(L)}}{\Delta\lambda_{1/2}^{(G)}} > 1` with

    `a` = [1., 0.15882, 1.04388, -1.38281, 0.46251, 0.82325, -0.58026] and

    `b` = [1., 0, 0.57575, 0.37902, -0.42519, -0.31525, 0.31718].

    While the :math:`\eta` function is fitted as:
    :math:`\eta=exp(\sum_{n=0}^{5}c_n(ln(\frac{\Delta\lambda_{1/2}^{(L)}}{\Delta\lambda_{1/2}^{(V)}}))^n`
    for :math:`0.01<\frac{\Delta\lambda_{1/2}^{(L)}}{\Delta\lambda_{1/2}^{(V)}}<0.999` with

    `c` = [5.14820e-04, 1.38821e+00, -9.60424e-02, -3.83995e-02, -7.40042e-03, -5.47626e-04].

    :param Line line: The emission line object for this line shape.
    :param float wavelength: The rest wavelength for this emission line.
    :param Species target_species: The target plasma species that is emitting.
    :param Plasma plasma: The emitting plasma object.
    :param AtomicData atomic_data: The atomic data provider.
    :param tuple stark_model_coefficients: Stark model coefficients in the form (c_ij, a_ij, b_ij).
                                           Default is None (will use
                                           `atomic_data.stark_model_coefficients`).
    :param Integrator1D integrator: Integrator1D instance to integrate the line shape
        over the spectral bin. Default is `GaussianQuadrature()`.
    :param str polarisation: Leaves only :math:`\pi`-/:math:`\sigma`-polarised components:
                             "pi" - leave only :math:`\pi`-polarised components,
                             "sigma" - leave only :math:`\sigma`-polarised components,
                             "no" - leave all components (default).
    """

    def __init__(self, Line line, double wavelength, Species target_species, Plasma plasma, AtomicData atomic_data,
                 tuple stark_model_coefficients=None, Integrator1D integrator=GaussianQuadrature(), polarisation='no'):

        super().__init__(line, wavelength, target_species, plasma, atomic_data, polarisation, integrator)

        try:
            # Fitted Stark Constants
            cij, aij, bij = stark_model_coefficients or self.atomic_data.stark_model_coefficients(line)
            if cij <= 0:
                raise ValueError('Coefficient c_ij must be positive.')
            if aij <= 0:
                raise ValueError('Coefficient a_ij must be positive.')
            if bij <= 0:
                raise ValueError('Coefficient b_ij must be positive.')
            self._aij = aij
            self._bij = bij
            self._cij = cij
        except IndexError:
            raise ValueError('Stark broadening coefficients for {} is not currently available.'.format(line))

        # polynomial coefficients in increasing powers
        self._fwhm_poly_coeff_gauss = [1., 0, 0.57575, 0.37902, -0.42519, -0.31525, 0.31718]
        self._fwhm_poly_coeff_lorentz = [1., 0.15882, 1.04388, -1.38281, 0.46251, 0.82325, -0.58026]

        self._weight_poly_coeff = [5.14820e-04, 1.38821e+00, -9.60424e-02, -3.83995e-02, -7.40042e-03, -5.47626e-04]

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.initializedcheck(False)
    @cython.cdivision(True)
    cpdef Spectrum add_line(self, double radiance, Point3D point, Vector3D direction, Spectrum spectrum):

        cdef:
            double ne, te, ts, shifted_wavelength, photon_energy, b_magn, comp_radiance, cos_sqr, sin_sqr
            double sigma, fwhm_lorentz, fwhm_gauss, fwhm_full, fwhm_ratio, fwhm_lorentz_to_total, lorentz_weight, gauss_weight
            cdef Vector3D ion_velocity, b_field
            int i

        ne = self.plasma.get_electron_distribution().density(point.x, point.y, point.z)

        te = self.plasma.get_electron_distribution().effective_temperature(point.x, point.y, point.z)

        fwhm_lorentz = self._cij * ne**self._aij / (te**self._bij) if ne > 0 and te > 0 else 0

        ts = self.target_species.distribution.effective_temperature(point.x, point.y, point.z)

        fwhm_gauss = _SIGMA2FWHM * thermal_broadening(self.wavelength, ts, self.line.element.atomic_weight) if ts > 0 else 0

        if fwhm_lorentz == 0 and fwhm_gauss == 0:
            return spectrum

        # calculating full FWHM
        if fwhm_gauss <= fwhm_lorentz:
            fwhm_ratio = fwhm_gauss / fwhm_lorentz
            fwhm_full = self._fwhm_poly_coeff_gauss[0]
            for i in range(1, 7):
                fwhm_full += self._fwhm_poly_coeff_gauss[i] * fwhm_ratio**i
            fwhm_full *= fwhm_lorentz
        else:
            fwhm_ratio = fwhm_lorentz / fwhm_gauss
            fwhm_full = self._fwhm_poly_coeff_lorentz[0]
            for i in range(1, 7):
                fwhm_full += self._fwhm_poly_coeff_lorentz[i] * fwhm_ratio**i
            fwhm_full *= fwhm_gauss

        sigma = fwhm_full / _SIGMA2FWHM

        fwhm_lorentz_to_total = fwhm_lorentz / fwhm_full

        # calculating Lorentzian weight
        if fwhm_lorentz_to_total < 0.01:
            lorentz_weight = 0
            fwhm_full = 0  # force add_lorentzian_line() to immediately return
        elif fwhm_lorentz_to_total > 0.999:
            lorentz_weight = 1
            sigma = 0  # force add_gaussian_line() to immediately return
        else:
            lorentz_weight = self._weight_poly_coeff[0]
            for i in range(1, 6):
                lorentz_weight += self._weight_poly_coeff[i] * log(fwhm_lorentz_to_total)**i
            lorentz_weight = exp(lorentz_weight)

        gauss_weight = 1 - lorentz_weight

        ion_velocity = self.target_species.distribution.bulk_velocity(point.x, point.y, point.z)

        # calculate emission line central wavelength, doppler shifted along observation direction
        shifted_wavelength = doppler_shift(self.wavelength, direction, ion_velocity)

        # obtain magnetic field
        b_field = self.plasma.get_b_field().evaluate(point.x, point.y, point.z)
        b_magn = b_field.get_length()

        if b_magn == 0:
            # no splitting if magnetic field strength is zero
            if self._polarisation != NO_POLARISATION:
                radiance *= 0.5  # pi or sigma polarisation, collecting only half of intensity

            spectrum = add_gaussian_line(gauss_weight * radiance, shifted_wavelength, sigma, spectrum)
            spectrum = add_lorentzian_line(lorentz_weight * radiance, shifted_wavelength, fwhm_full, spectrum, self.integrator)

            return spectrum

        # coefficients for intensities parallel and perpendicular to magnetic field
        cos_sqr = (b_field.dot(direction.normalise()) / b_magn)**2
        sin_sqr = 1. - cos_sqr

        # adding pi component of the Zeeman triplet in case of NO_POLARISATION or PI_POLARISATION
        if self._polarisation != SIGMA_POLARISATION:
            comp_radiance = 0.5 * sin_sqr * radiance
            spectrum = add_gaussian_line(gauss_weight * comp_radiance, shifted_wavelength, sigma, spectrum)
            spectrum = add_lorentzian_line(lorentz_weight * comp_radiance, shifted_wavelength, fwhm_full, spectrum, self.integrator)

        # adding sigma +/- components of the Zeeman triplet in case of NO_POLARISATION or SIGMA_POLARISATION
        if self._polarisation != PI_POLARISATION:
            comp_radiance = (0.25 * sin_sqr + 0.5 * cos_sqr) * radiance

            photon_energy = HC_EV_NM / self.wavelength

            shifted_wavelength = doppler_shift(HC_EV_NM / (photon_energy - BOHR_MAGNETON * b_magn), direction, ion_velocity)
            spectrum = add_gaussian_line(gauss_weight * comp_radiance, shifted_wavelength, sigma, spectrum)
            spectrum = add_lorentzian_line(lorentz_weight * comp_radiance, shifted_wavelength, fwhm_full, spectrum, self.integrator)

            shifted_wavelength = doppler_shift(HC_EV_NM / (photon_energy + BOHR_MAGNETON * b_magn), direction, ion_velocity)
            spectrum = add_gaussian_line(gauss_weight * comp_radiance, shifted_wavelength, sigma, spectrum)
            spectrum = add_lorentzian_line(lorentz_weight * comp_radiance, shifted_wavelength, fwhm_full, spectrum, self.integrator)

        return spectrum




# Copyright 2016-2021 Euratom
# Copyright 2016-2021 United Kingdom Atomic Energy Authority
# Copyright 2016-2021 Centro de Investigaciones Energéticas, Medioambientales y Tecnológicas
#
# Licensed under the EUPL, Version 1.1 or – as soon they will be approved by the
# European Commission - subsequent versions of the EUPL (the "Licence");
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at:
#
# https://joinup.ec.europa.eu/software/page/eupl5
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the Licence is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.
#
# See the Licence for the specific language governing permissions and limitations
# under the Licence.


import unittest
import os
from collections.abc import Iterable
import numpy as np
from raysect.core.math.function.float import Function1D, Function2D, Interpolator1DArray, Interpolator2DArray

from cherab.core.atomic import neon, hydrogen, helium
from cherab.core.math import AxisymmetricMapper
from cherab.openadas import OpenADAS
from cherab.tools.plasmas.ionisation_balance import (fractional_abundance, from_elementdensity, match_plasma_neutrality,
                                                     interpolators1d_fractional, interpolators1d_from_elementdensity,
                                                     interpolators1d_match_plasma_neutrality,
                                                     interpolators2d_fractional, interpolators2d_from_elementdensity,
                                                     interpolators2d_match_plasma_neutrality,
                                                     abundance_axisymmetric_mapper,
                                                     equilibrium_map3d_fractional, equilibrium_map3d_from_elementdensity,
                                                     equilibrium_map3d_match_plasma_neutrality)

from cherab.tools.equilibrium import example_equilibrium


def double_parabola(r, centre, edge, p, q):
    return (centre - edge) * np.power((1 - np.power((r - r.min()) / (r.max() - r.min()), p)), q) + edge


def normal(x, mu, sd, height=1, offset=0):
    return height * np.exp(-1 * np.power(x - mu, 2) / (2 * sd ** 2)) + offset


def get_electron_density_profile(abundances):
    n_e = np.zeros((abundances[0].shape[1]))
    for abundance in abundances:
        for rownumber, row in enumerate(abundance.T):
            n_e[rownumber] += np.sum(row * np.arange(row.shape[0]))

    return n_e


def get_electron_density_spot(densities):
    n_e = 0
    for spec in densities:
        for index, value in enumerate(spec):
            n_e += index * value

    return n_e


def exp_decay(r, lamb, max_val):
    return max_val * np.exp((r - r.max()) * lamb)


class TestIonizationBalance1D(unittest.TestCase):

    # create plasma profiles and interpolators
    # 1d profiles
    psin_1d = np.linspace(0, 1.1, 15, endpoint=True)
    psin_1d_detailed = np.linspace(0, 1.1, 50, endpoint=True)

    t_e_profile_1d = double_parabola(psin_1d, 5000, 10, 2, 2)
    n_e_profile_1d = double_parabola(psin_1d, 6e19, 5e18, 2, 2)

    t_element_profile_1d = double_parabola(psin_1d, 1500, 40, 2, 2)
    n_element_profile_1d = double_parabola(psin_1d, 1e17, 1e17, 2, 2) + normal(psin_1d, 0.9, 0.1, 5e17)
    n_element2_profile_1d = double_parabola(psin_1d, 5e17, 1e17, 2, 2)

    n_tcx_donor_profile_1d = exp_decay(psin_1d, 10, 3e16)

    t_e_1d = Interpolator1DArray(psin_1d, t_e_profile_1d, 'cubic', 'none', 0)
    n_e_1d = Interpolator1DArray(psin_1d, n_e_profile_1d, 'cubic', 'none', 0)

    t_element_1d = Interpolator1DArray(psin_1d, t_element_profile_1d, 'cubic', 'none', 0)
    n_element_1d = Interpolator1DArray(psin_1d, n_element_profile_1d, 'cubic', 'none', 0)
    n_element2_1d = Interpolator1DArray(psin_1d, n_element2_profile_1d, 'cubic', 'none', 0)

    n_tcx_donor_1d = Interpolator1DArray(psin_1d, n_tcx_donor_profile_1d, 'cubic', 'none', 0)

    # denser psi array to test interpolators
    n_e_profile_detailed_1d = np.zeros_like(psin_1d_detailed)
    for index, value in enumerate(psin_1d_detailed):
        n_e_profile_detailed_1d[index] = n_e_1d(value)

    # 2d profiles

    r = np.linspace(-1, 1, 6)
    z = np.linspace(-1, 1, 8)

    psin_2d = np.zeros((*r.shape, *z.shape))

    for index0, value0 in enumerate(r):
        for index1, value1 in enumerate(z):
            if np.sqrt(value0 ** 2 + value1 ** 2) < psin_1d.max():
                psin_2d[index0, index1] = np.sqrt(value0 ** 2 + value1 ** 2)
            else:
                psin_2d[index0, index1] = psin_1d.max()

    t_e_profile_2d = np.zeros_like(psin_2d)
    for index in np.ndindex(*t_e_profile_2d.shape):
        t_e_profile_2d[index] = t_e_1d(psin_2d[index])

    t_e_2d = Interpolator2DArray(r, z, t_e_profile_2d, 'cubic', 'none', 0, 0)

    n_e_profile_2d = np.zeros_like(psin_2d)
    for index in np.ndindex(*n_e_profile_2d.shape):
        n_e_profile_2d[index] = n_e_1d(psin_2d[index])

    n_e_2d = Interpolator2DArray(r, z, n_e_profile_2d, 'cubic', 'none', 0, 0)

    t_element_profile_2d = np.zeros_like(psin_2d)
    for index in np.ndindex(*t_element_profile_2d.shape):
        t_element_profile_2d[index] = t_element_1d(psin_2d[index])

    t_element_2d = Interpolator2DArray(r, z, t_element_profile_2d, 'cubic', 'none', 0, 0)

    n_element_profile_2d = np.zeros_like(psin_2d)
    for index in np.ndindex(*n_element_profile_2d.shape):
        n_element_profile_2d[index] = n_element_1d(psin_2d[index])

    n_element_2d = Interpolator2DArray(r, z, n_element_profile_2d, 'cubic', 'none', 0, 0)

    n_element2_profile_2d = np.zeros_like(psin_2d)
    for index in np.ndindex(*n_element2_profile_2d.shape):
        n_element2_profile_2d[index] = n_element2_1d(psin_2d[index])

    n_element2_2d = Interpolator2DArray(r, z, n_element2_profile_2d, 'cubic', 'none', 0, 0)

    n_tcx_donor_profile_2d = np.zeros_like(psin_2d)
    for index in np.ndindex(*n_tcx_donor_profile_2d.shape):
        n_tcx_donor_profile_2d[index] = n_tcx_donor_1d(psin_2d[index])

    n_tcx_donor_2d = Interpolator2DArray(r, z, n_element_profile_2d, 'cubic', 'none', 0, 0)

    # define psi for single-point tests
    psi_value = 0.9

    # load adas atomic database and define elements
    repository_path = os.path.join(os.path.dirname(__file__), 'data/atomic_rates_mockup')
    atomic_data = OpenADAS(data_path=repository_path, permit_extrapolation=True)

    element = neon
    element2 = helium
    element_bulk = hydrogen

    tcx_donor = hydrogen

    TOLERANCE = 1e-3

    def sumup_fractions(self, fractions):

        if isinstance(fractions, dict):
            if isinstance(fractions[0], Iterable):
                total = np.zeros_like(fractions[0])
            else:
                total = 0

            for index, values in fractions.items():
                total += values

        elif isinstance(fractions, np.ndarray):
            total = np.zeros_like(fractions[0, ...])

            for index in np.ndindex(fractions.shape):
                total[index] += fractions[index]

        return total

    def sumup_electrons(self, densities):
        if isinstance(densities, dict):
            if isinstance(densities[0], Iterable):
                total = np.zeros_like(densities[0])
            else:
                total = 0

            for index, values in densities.items():
                total += values * index

        elif isinstance(densities, np.ndarray):
            total = np.zeros_like(densities[0, ...])

            for index in np.ndindex(densities.shape):
                total[index] += densities[index] * index[0]

        return total

    def evaluate_interpolators(self, interpolators, free_variable):

        profiles = {}
        for key, item in interpolators.items():
            if isinstance(item, Function1D):
                profiles[key] = np.zeros_like(free_variable)
                for index in np.ndindex(*free_variable.shape):
                    profiles[key][index] = item(free_variable[index])
            elif isinstance(item, Function2D):
                profiles[key] = np.zeros((*free_variable[0].shape, *free_variable[1].shape))
                for index0, value0 in enumerate(free_variable[0]):
                    for index1, value1 in enumerate(free_variable[1]):
                        profiles[key][index0, index1] = item(value0, value1)
            elif isinstance(item, AxisymmetricMapper):
                profiles[key] = np.zeros_like(free_variable)
                for index, value in enumerate(free_variable):
                    profiles[key][index] = item(value, 0, 0)

        return profiles

    def test_fractional_0d_from_0d(self):
        """
        test fractional abundance calculation with float numbers as inputs
        """
        abundance_fractional = fractional_abundance(self.atomic_data, self.element, self.n_e_1d(self.psi_value),
                                                    self.t_e_1d(self.psi_value))

        fraction_sum = self.sumup_fractions(abundance_fractional)
        self.assertTrue(1 - fraction_sum < self.TOLERANCE)

    def test_fractional_0d_from_0d_tcx(self):
        """
        test fractional abundance calculation with thermal cx and float numbers as inputs
        """
        abundance_fractional = fractional_abundance(self.atomic_data, self.element, self.n_e_1d(self.psi_value),
                                                    self.t_e_1d(self.psi_value),
                                                    self.tcx_donor, self.n_tcx_donor_1d(self.psi_value), 0)

        fraction_sum = self.sumup_fractions(abundance_fractional)
        self.assertTrue(1 - fraction_sum < self.TOLERANCE)

    def test_fractional_0d_from_interpolators(self):
        """
        test interpolators and free_variable as inputs
        """

        abundance_fractional = fractional_abundance(self.atomic_data, self.element, self.n_e_1d, self.t_e_1d,
                                                    free_variable=self.psi_value)

        fraction_sum = self.sumup_fractions(abundance_fractional)
        self.assertTrue(1 - fraction_sum < self.TOLERANCE)

    def test_fractional_0d_from_interpolators_tcx(self):
        """
        test interpolators and free_variable as inputs
        """

        abundance_fractional = fractional_abundance(self.atomic_data, self.element, self.n_e_1d, self.t_e_1d,
                                                    tcx_donor=self.tcx_donor, tcx_donor_n=self.n_tcx_donor_1d,
                                                    tcx_donor_charge=0, free_variable=self.psi_value)

        fraction_sum = self.sumup_fractions(abundance_fractional)
        self.assertTrue(1 - fraction_sum < self.TOLERANCE)

    def test_fractional_0d_from_mixed(self):
        """
        test mixed types of inputs
        """

        abundance_fractional = fractional_abundance(self.atomic_data, self.element, self.n_e_1d(self.psi_value),
                                                    self.t_e_1d,
                                                    free_variable=self.psi_value)

        fraction_sum = self.sumup_fractions(abundance_fractional)
        self.assertTrue(1 - fraction_sum < self.TOLERANCE)

        abundance_fractional = fractional_abundance(self.atomic_data, self.element, self.n_e_1d,
                                                    self.t_e_1d(self.psi_value),
                                                    free_variable=self.psi_value)

        fraction_sum = self.sumup_fractions(abundance_fractional)
        self.assertTrue(1 - fraction_sum < self.TOLERANCE)

    def test_fractional_1d_from_1d(self):
        """
        test calculation of 1d fractional profiles with 1d iterables as inputs
        """
        abundance_fractional = fractional_abundance(self.atomic_data, self.element, self.n_e_profile_1d,
                                                    self.t_e_profile_1d)

        fraction_sum = self.sumup_fractions(abundance_fractional)
        self.assertTrue(np.allclose(fraction_sum, 1, atol=self.TOLERANCE))

    def test_fractional_1d_from_1d_tcx(self):
        """
        test calculation of 1d fractional profiles with 1d iterables as inputs
        """

        abundance_fractional = fractional_abundance(self.atomic_data, self.element, self.n_e_profile_1d,
                                                    self.t_e_profile_1d,
                                                    tcx_donor=self.tcx_donor, tcx_donor_n=self.n_tcx_donor_profile_1d,
                                                    tcx_donor_charge=0)

        fraction_sum = self.sumup_fractions(abundance_fractional)
        self.assertTrue(np.allclose(fraction_sum, 1, atol=self.TOLERANCE))

    def test_fractional_1d_from_interpolators(self):
        """
        test calculation of 1d fractional profiles with 1d interpolators as inputs
        """
        abundance_fractional = fractional_abundance(self.atomic_data, self.element, self.n_e_1d, self.t_e_1d,
                                                    free_variable=self.psin_1d)

        fraction_sum = self.sumup_fractions(abundance_fractional)
        self.assertTrue(np.allclose(fraction_sum, 1, atol=self.TOLERANCE))

    def test_fractional_1d_from_interpolators_tcx(self):
        """
        test calculation of 1d fractional profiles with 1d iterables as inputs
        """

        abundance_fractional = fractional_abundance(self.atomic_data, self.element, self.n_e_1d, self.t_e_1d,
                                                    tcx_donor=self.tcx_donor, tcx_donor_n=self.n_tcx_donor_1d,
                                                    tcx_donor_charge=0, free_variable=self.psin_1d)

        fraction_sum = self.sumup_fractions(abundance_fractional)
        self.assertTrue(np.allclose(fraction_sum, 1, atol=self.TOLERANCE))

    def test_fractional_from_mixed(self):
        """
        test calculation of 1d fractional profiles with mixed types as inputs
        """

        abundance_fractional = fractional_abundance(self.atomic_data, self.element, self.n_e_profile_1d, self.t_e_1d,
                                                    free_variable=self.psin_1d)

        fraction_sum = self.sumup_fractions(abundance_fractional)
        self.assertTrue(np.allclose(fraction_sum, 1, atol=self.TOLERANCE))

        abundance_fractional = fractional_abundance(self.atomic_data, self.element, self.n_e_1d, self.t_e_profile_1d,
                                                    free_variable=self.psin_1d)

        fraction_sum = self.sumup_fractions(abundance_fractional)
        self.assertTrue(np.allclose(fraction_sum, 1, atol=self.TOLERANCE))

    def test_fractional_from_mixed_tcx(self):
        """
        test calculation of 1d fractional profiles with mixed types as inputs
        """

        abundance_fractional = fractional_abundance(self.atomic_data, self.element, self.n_e_profile_1d, self.t_e_1d,
                                                    tcx_donor=self.tcx_donor, tcx_donor_n=self.n_tcx_donor_1d,
                                                    tcx_donor_charge=0, free_variable=self.psin_1d)

        fraction_sum = self.sumup_fractions(abundance_fractional)
        self.assertTrue(np.allclose(fraction_sum, 1, atol=self.TOLERANCE))

        abundance_fractional = fractional_abundance(self.atomic_data, self.element, self.n_e_1d, self.t_e_profile_1d,
                                                    tcx_donor=self.tcx_donor, tcx_donor_n=self.n_tcx_donor_1d,
                                                    tcx_donor_charge=0, free_variable=self.psin_1d)

        fraction_sum = self.sumup_fractions(abundance_fractional)
        self.assertTrue(np.allclose(fraction_sum, 1, atol=self.TOLERANCE))

        abundance_fractional = fractional_abundance(self.atomic_data, self.element, self.n_e_1d, self.t_e_profile_1d,
                                                    tcx_donor=self.tcx_donor, tcx_donor_n=self.n_tcx_donor_profile_1d,
                                                    tcx_donor_charge=0, free_variable=self.psin_1d)

        fraction_sum = self.sumup_fractions(abundance_fractional)
        self.assertTrue(np.allclose(fraction_sum, 1, atol=self.TOLERANCE))

    def test_fractional_inetrpolators_1d(self):
        """
        test calculation of 1d fractional interpolators
        """

        interpolators_fractional = interpolators1d_fractional(self.atomic_data, self.element, self.psin_1d,
                                                              self.n_e_profile_1d, self.t_e_1d)

        profiles = self.evaluate_interpolators(interpolators_fractional, self.psin_1d)
        fraction_sum = self.sumup_fractions(profiles)

        self.assertTrue(np.allclose(fraction_sum, 1, atol=self.TOLERANCE))

    def test_fractional_inetrpolators_1d_tcx(self):
        """
        test calculation of 1d fractional interpolators with thermal cx
        """

        interpolators_fractional = interpolators1d_fractional(self.atomic_data, self.element, self.psin_1d,
                                                              self.n_e_profile_1d, self.t_e_1d,
                                                              tcx_donor=self.tcx_donor,
                                                              tcx_donor_n=self.n_tcx_donor_profile_1d,
                                                              tcx_donor_charge=0)

        profiles = self.evaluate_interpolators(interpolators_fractional, self.psin_1d)
        fraction_sum = self.sumup_fractions(profiles)

        self.assertTrue(np.allclose(fraction_sum, 1, atol=self.TOLERANCE))

    def test_balance_0d_elementdensity(self):
        """
        test calculation of ionization balance
        """
        # test with floats as input
        densities = from_elementdensity(self.atomic_data, self.element, self.n_element_1d(self.psi_value),
                                        self.n_e_1d(self.psi_value), self.t_e_1d(self.psi_value))
        total = self.sumup_fractions(densities)
        self.assertTrue(np.isclose(total, self.n_element_1d(self.psi_value), rtol=self.TOLERANCE))

        # test with interpolators
        densities = from_elementdensity(self.atomic_data, self.element, self.n_element_1d(self.psi_value),
                                        self.n_e_1d(self.psi_value), self.t_e_1d(self.psi_value))
        total = self.sumup_fractions(densities)
        self.assertTrue(np.isclose(total, self.n_element_1d(self.psi_value), rtol=self.TOLERANCE))

        # test with mixed parameters
        densities = from_elementdensity(self.atomic_data, self.element, self.n_element_1d,
                                        self.n_e_1d(self.psi_value), self.t_e_1d(self.psi_value),
                                        free_variable=self.psi_value)
        total = self.sumup_fractions(densities)
        self.assertTrue(np.isclose(total, self.n_element_1d(self.psi_value), rtol=self.TOLERANCE))

    def test_balance_0d_plasma_neutrality(self):
        """test matching of plasma neutrality"""

        densities_1 = from_elementdensity(self.atomic_data, self.element, self.n_element_1d,
                                          self.n_e_1d(self.psi_value), self.t_e_1d(self.psi_value),
                                          free_variable=self.psi_value)

        densities_2 = from_elementdensity(self.atomic_data, self.element2, self.n_element2_1d,
                                          self.n_e_1d(self.psi_value), self.t_e_1d(self.psi_value),
                                          free_variable=self.psi_value)

        densities_3 = match_plasma_neutrality(self.atomic_data, self.element_bulk, [densities_1, densities_2],
                                              self.n_e_1d(self.psi_value), self.t_e_1d(self.psi_value))

        total = self.sumup_electrons(densities_1)
        total += self.sumup_electrons(densities_2)
        total += self.sumup_electrons(densities_3)

        self.assertTrue(np.isclose(total, self.n_e_1d(self.psi_value), rtol=self.TOLERANCE))

    def test_balance_0d_plasma_neutrality_tcx(self):
        """test matching of plasma neutrality"""

        densities_1 = from_elementdensity(self.atomic_data, self.element, self.n_element_1d,
                                          self.n_e_1d(self.psi_value), self.t_e_1d(self.psi_value),
                                          free_variable=self.psi_value,
                                          tcx_donor=self.tcx_donor, tcx_donor_n=self.n_tcx_donor_1d, tcx_donor_charge=0,
                                          )

        densities_2 = from_elementdensity(self.atomic_data, self.element2, self.n_element2_1d,
                                          self.n_e_1d(self.psi_value), self.t_e_1d(self.psi_value),
                                          free_variable=self.psi_value,
                                          tcx_donor=self.tcx_donor, tcx_donor_n=self.n_tcx_donor_1d, tcx_donor_charge=0,
                                          )

        densities_3 = match_plasma_neutrality(self.atomic_data, self.element_bulk, [densities_1, densities_2],
                                              self.n_e_1d(self.psi_value), self.t_e_1d(self.psi_value))

        total = self.sumup_electrons(densities_1)
        total += self.sumup_electrons(densities_2)
        total += self.sumup_electrons(densities_3)

        self.assertTrue(np.isclose(total, self.n_e_1d(self.psi_value), rtol=self.TOLERANCE))

    def test_balance_1d_plasma_neutrality(self):
        """test matching of plasma neutrality for 1d profiles"""

        densities_1 = from_elementdensity(self.atomic_data, self.element, self.n_element_1d,
                                          self.n_e_1d, self.t_e_profile_1d,
                                          free_variable=self.psin_1d)

        densities_2 = from_elementdensity(self.atomic_data, self.element2, self.n_element2_1d,
                                          self.n_e_profile_1d, self.t_e_1d,
                                          free_variable=self.psin_1d)

        densities_3 = match_plasma_neutrality(self.atomic_data, self.element_bulk, [densities_1, densities_2],
                                              self.n_e_1d, self.t_e_profile_1d,
                                              free_variable=self.psin_1d)

        total = self.sumup_electrons(densities_1)
        total += self.sumup_electrons(densities_2)
        total += self.sumup_electrons(densities_3)

        self.assertTrue(np.allclose(total, self.n_e_profile_1d, rtol=self.TOLERANCE))

    def test_balance_1d_plasma_neutrality_tcx(self):
        """test matching of plasma neutrality for 1d profiles with thermal cx"""

        densities_1 = from_elementdensity(self.atomic_data, self.element, self.n_element_1d,
                                          self.n_e_1d, self.t_e_profile_1d,
                                          free_variable=self.psin_1d,
                                          tcx_donor=self.tcx_donor, tcx_donor_n=self.n_tcx_donor_1d, tcx_donor_charge=0
                                          )

        densities_2 = from_elementdensity(self.atomic_data, self.element2, self.n_element2_1d,
                                          self.n_e_profile_1d, self.t_e_1d,
                                          free_variable=self.psin_1d,
                                          tcx_donor=self.tcx_donor, tcx_donor_n=self.n_tcx_donor_1d, tcx_donor_charge=0)

        densities_3 = match_plasma_neutrality(self.atomic_data, self.element_bulk, [densities_1, densities_2],
                                              self.n_e_1d, self.t_e_profile_1d,
                                              free_variable=self.psin_1d,
                                              tcx_donor=self.tcx_donor, tcx_donor_n=self.n_tcx_donor_1d,
                                              tcx_donor_charge=0)

        total = self.sumup_electrons(densities_1)
        total += self.sumup_electrons(densities_2)
        total += self.sumup_electrons(densities_3)

        self.assertTrue(np.allclose(total, self.n_e_profile_1d, rtol=self.TOLERANCE))

    def test_balance_1d_interpolators_from_element_density(self):
        """
        test calculation of 1d interpolators of charge stage densities
        """

        interpolators_abundance = interpolators1d_from_elementdensity(self.atomic_data, self.element, self.psin_1d,
                                                                      self.n_element_1d,
                                                                      self.n_e_profile_1d, self.t_e_1d)

        profiles = self.evaluate_interpolators(interpolators_abundance, self.psin_1d)
        fraction_sum = self.sumup_fractions(profiles)

        self.assertTrue(np.allclose(fraction_sum, self.n_element_profile_1d, rtol=self.TOLERANCE))

    def test_balance_1d_interpolators_from_element_density_tcx(self):
        """
        test calculation of 1d interpolators of ion charge state densities
        """

        interpolators_abundance = interpolators1d_from_elementdensity(self.atomic_data, self.element, self.psin_1d,
                                                                      self.n_element_1d,
                                                                      self.n_e_profile_1d, self.t_e_1d,
                                                                      tcx_donor=self.tcx_donor,
                                                                      tcx_donor_n=self.n_tcx_donor_1d,
                                                                      tcx_donor_charge=0)

        profiles = self.evaluate_interpolators(interpolators_abundance, self.psin_1d)
        fraction_sum = self.sumup_fractions(profiles)

        self.assertTrue(np.allclose(fraction_sum, self.n_element_profile_1d, rtol=self.TOLERANCE))

    def test_balance_1d_interpolators_plasma_neutrality(self):
        """
        test calulation of 1d interpolators for ion charge state densities using plasma neutrality condition.
        """

        interpolators_abundance_1 = interpolators1d_from_elementdensity(self.atomic_data, self.element, self.psin_1d,
                                                                        self.n_element_1d,
                                                                        self.n_e_profile_1d, self.t_e_1d)

        interpolators_abundance_2 = interpolators1d_from_elementdensity(self.atomic_data, self.element2, self.psin_1d,
                                                                        self.n_element2_1d,
                                                                        self.n_e_profile_1d, self.t_e_1d)

        interpolators_abundance_3 = interpolators1d_match_plasma_neutrality(self.atomic_data, self.element,
                                                                            self.psin_1d,
                                                                            [interpolators_abundance_1,
                                                                             interpolators_abundance_2],
                                                                            self.n_e_profile_1d, self.t_e_1d)

        profiles1 = self.evaluate_interpolators(interpolators_abundance_1, self.psin_1d)
        profiles2 = self.evaluate_interpolators(interpolators_abundance_2, self.psin_1d)
        profiles3 = self.evaluate_interpolators(interpolators_abundance_3, self.psin_1d)

        total = self.sumup_electrons(profiles1)
        total += self.sumup_electrons(profiles2)
        total += self.sumup_electrons(profiles3)

        self.assertTrue(np.allclose(total, self.n_e_profile_1d, rtol=self.TOLERANCE))

    def test_balance_1d_interpolators_plasma_neutrality_tcx(self):
        """
        test calulation of 1d interpolators for ion charge state densities using plasma neutrality condition.
        """

        interpolators_abundance_1 = interpolators1d_from_elementdensity(self.atomic_data, self.element, self.psin_1d,
                                                                        self.n_element_1d,
                                                                        self.n_e_profile_1d, self.t_e_1d,
                                                                        tcx_donor=self.tcx_donor,
                                                                        tcx_donor_n=self.n_tcx_donor_1d,
                                                                        tcx_donor_charge=0)

        interpolators_abundance_2 = interpolators1d_from_elementdensity(self.atomic_data, self.element2, self.psin_1d,
                                                                        self.n_element2_1d,
                                                                        self.n_e_profile_1d, self.t_e_1d,
                                                                        tcx_donor=self.tcx_donor,
                                                                        tcx_donor_n=self.n_tcx_donor_1d,
                                                                        tcx_donor_charge=0)

        interpolators_abundance_3 = interpolators1d_match_plasma_neutrality(self.atomic_data, self.element,
                                                                            self.psin_1d,
                                                                            [interpolators_abundance_1,
                                                                             interpolators_abundance_2],
                                                                            self.n_e_profile_1d, self.t_e_1d,
                                                                            tcx_donor=self.tcx_donor,
                                                                            tcx_donor_n=self.n_tcx_donor_1d,
                                                                            tcx_donor_charge=0)

        profiles1 = self.evaluate_interpolators(interpolators_abundance_1, self.psin_1d)
        profiles2 = self.evaluate_interpolators(interpolators_abundance_2, self.psin_1d)
        profiles3 = self.evaluate_interpolators(interpolators_abundance_3, self.psin_1d)

        total = self.sumup_electrons(profiles1)
        total += self.sumup_electrons(profiles2)
        total += self.sumup_electrons(profiles3)

        self.assertTrue(np.allclose(total, self.n_e_profile_1d, rtol=self.TOLERANCE))

    def test_fractional_2d_from_2d(self):
        """
        test fractional abundance 2d profile calculation with arrays as inputs
        """
        abundance_fractional = fractional_abundance(self.atomic_data, self.element, self.n_e_profile_2d,
                                                    self.t_e_profile_2d)

        fraction_sum = self.sumup_fractions(abundance_fractional)
        self.assertTrue(np.allclose(fraction_sum, np.ones_like(self.n_e_profile_2d), rtol=self.TOLERANCE))

    def test_fractional_2d_from_2d_tcx(self):
        """
        test fractional abundance 2d profile calculation with arrays as inputs with thermal cx
        """
        abundance_fractional = fractional_abundance(self.atomic_data, self.element, self.n_e_profile_2d,
                                                    self.t_e_profile_2d, tcx_donor=self.tcx_donor,
                                                    tcx_donor_n=self.n_tcx_donor_profile_2d, tcx_donor_charge=0)

        fraction_sum = self.sumup_fractions(abundance_fractional)
        self.assertTrue(np.allclose(fraction_sum, np.ones_like(self.n_e_profile_2d), rtol=self.TOLERANCE))

    def test_fractional_2d_from_interpolators(self):
        """
        test fractional abundance 2d profile calculation with interpolators as inputs
        """
        abundance_fractional = fractional_abundance(self.atomic_data, self.element, self.n_e_2d,
                                                    self.t_e_2d, free_variable=(self.r, self.z))

        fraction_sum = self.sumup_fractions(abundance_fractional)
        self.assertTrue(np.allclose(fraction_sum, np.ones_like(self.n_e_profile_2d), rtol=self.TOLERANCE))

    def test_fractional_2d_from_interpolators_tcx(self):
        """
        test fractional abundance 2d profile calculation with interpolators as inputs with thermal cx
        """
        abundance_fractional = fractional_abundance(self.atomic_data, self.element, self.n_e_2d,
                                                    self.t_e_2d, tcx_donor=self.tcx_donor,
                                                    tcx_donor_n=self.n_tcx_donor_2d, tcx_donor_charge=0,
                                                    free_variable=(self.r, self.z))

        fraction_sum = self.sumup_fractions(abundance_fractional)
        self.assertTrue(np.allclose(fraction_sum, np.ones_like(self.n_e_profile_2d), rtol=self.TOLERANCE))

    def test_fractional_2d_from_mixed(self):
        """
        test fractional abundance 2d profile calculation with mixed inputs
        """
        abundance_fractional = fractional_abundance(self.atomic_data, self.element, self.n_e_profile_2d,
                                                    self.t_e_2d, free_variable=(self.r, self.z))

        fraction_sum = self.sumup_fractions(abundance_fractional)
        self.assertTrue(np.allclose(fraction_sum, np.ones_like(self.n_e_profile_2d), rtol=self.TOLERANCE))

    def test_fractional_2d_from_mixed_tcx(self):
        """
        test fractional abundance 2d profile calculation with mixed inputs with thermal cx
        """
        abundance_fractional = fractional_abundance(self.atomic_data, self.element, self.n_e_profile_2d,
                                                    self.t_e_2d, tcx_donor=self.tcx_donor,
                                                    tcx_donor_n=self.n_tcx_donor_2d, tcx_donor_charge=0,
                                                    free_variable=(self.r, self.z))

        fraction_sum = self.sumup_fractions(abundance_fractional)
        self.assertTrue(np.allclose(fraction_sum, np.ones_like(self.n_e_profile_2d), rtol=self.TOLERANCE))

    def test_balance_2d_elementdensity(self):
        """
        test abundance 2d profile calculation from element density
        """
        abundance = from_elementdensity(self.atomic_data, self.element, self.n_element_profile_2d, self.n_e_2d,
                                        self.t_e_profile_2d, free_variable=(self.r, self.z))

        fraction_sum = self.sumup_fractions(abundance)
        self.assertTrue(np.allclose(fraction_sum, self.n_element_profile_2d, rtol=self.TOLERANCE))

    def test_balance_2d_elementdensity_tcx(self):
        """
        test abundance 2d profile calculation from element density
        """
        abundance = from_elementdensity(self.atomic_data, self.element, self.n_element_profile_2d, self.n_e_2d,
                                        self.t_e_profile_2d, tcx_donor=self.tcx_donor,
                                        tcx_donor_n=self.n_tcx_donor_2d, tcx_donor_charge=0,
                                        free_variable=(self.r, self.z))

        fraction_sum = self.sumup_fractions(abundance)
        self.assertTrue(np.allclose(fraction_sum, self.n_element_profile_2d, rtol=self.TOLERANCE))

    def test_balance_2d_plasma_neutrality(self):
        """test matching of plasma neutrality for 2d profiles"""

        densities_1 = from_elementdensity(self.atomic_data, self.element, self.n_element_2d,
                                          self.n_e_2d, self.t_e_profile_2d,
                                          free_variable=(self.r, self.z))

        densities_2 = from_elementdensity(self.atomic_data, self.element2, self.n_element2_2d,
                                          self.n_e_profile_2d, self.t_e_2d,
                                          free_variable=(self.r, self.z))

        densities_3 = match_plasma_neutrality(self.atomic_data, self.element_bulk, [densities_1, densities_2],
                                              self.n_e_2d, self.t_e_profile_2d,
                                              free_variable=(self.r, self.z))

        total = self.sumup_electrons(densities_1)
        total += self.sumup_electrons(densities_2)
        total += self.sumup_electrons(densities_3)

        self.assertTrue(np.allclose(total, self.n_e_profile_2d, rtol=self.TOLERANCE))

    def test_balance_2d_plasma_neutrality_tcx(self):
        """test matching of plasma neutrality for 2d profiles"""

        densities_1 = from_elementdensity(self.atomic_data, self.element, self.n_element_2d,
                                          self.n_e_2d, self.t_e_profile_2d, tcx_donor=self.tcx_donor,
                                          tcx_donor_n=self.n_tcx_donor_2d, tcx_donor_charge=0,
                                          free_variable=(self.r, self.z))

        densities_2 = from_elementdensity(self.atomic_data, self.element2, self.n_element2_2d,
                                          self.n_e_profile_2d, self.t_e_2d, tcx_donor=self.tcx_donor,
                                          tcx_donor_n=self.n_tcx_donor_2d, tcx_donor_charge=0,
                                          free_variable=(self.r, self.z))

        densities_3 = match_plasma_neutrality(self.atomic_data, self.element_bulk, [densities_1, densities_2],
                                              self.n_e_2d, self.t_e_profile_2d, tcx_donor=self.tcx_donor,
                                              tcx_donor_n=self.n_tcx_donor_2d, tcx_donor_charge=0,
                                              free_variable=(self.r, self.z))

        total = self.sumup_electrons(densities_1)
        total += self.sumup_electrons(densities_2)
        total += self.sumup_electrons(densities_3)

        self.assertTrue(np.allclose(total, self.n_e_profile_2d, rtol=self.TOLERANCE))

    def test_fractional_inetrpolators_2d(self):
        """
        test calculation of 1d fractional interpolators
        """

        interpolators_fractional = interpolators2d_fractional(self.atomic_data, self.element, (self.r, self.z),
                                                              self.n_e_profile_2d, self.t_e_2d)

        profiles = self.evaluate_interpolators(interpolators_fractional, (self.r, self.z))
        fraction_sum = self.sumup_fractions(profiles)

        self.assertTrue(np.allclose(fraction_sum, 1, atol=self.TOLERANCE))

    def test_fractional_inetrpolators_2d_tcx(self):
        """
        test calculation of 1d fractional interpolators with thermal cx
        """

        interpolators_fractional = interpolators2d_fractional(self.atomic_data, self.element, (self.r, self.z),
                                                              self.n_e_profile_2d, self.t_e_2d,
                                                              tcx_donor=self.tcx_donor,
                                                              tcx_donor_n=self.n_tcx_donor_profile_2d,
                                                              tcx_donor_charge=0)

        profiles = self.evaluate_interpolators(interpolators_fractional, (self.r, self.z))
        fraction_sum = self.sumup_fractions(profiles)

        self.assertTrue(np.allclose(fraction_sum, 1, atol=self.TOLERANCE))

    def test_balance_2d_interpolators_from_element_density(self):
        """
        test calculation of 2d interpolators of charge stage densities
        """

        interpolators_abundance = interpolators2d_from_elementdensity(self.atomic_data, self.element, (self.r, self.z),
                                                                      self.n_element_2d,
                                                                      self.n_e_profile_2d, self.t_e_2d)

        profiles = self.evaluate_interpolators(interpolators_abundance, (self.r, self.z))
        fraction_sum = self.sumup_fractions(profiles)

        self.assertTrue(np.allclose(fraction_sum, self.n_element_profile_2d, rtol=self.TOLERANCE))

    def test_balance_2d_interpolators_from_element_density_tcx(self):
        """
        test calculation of 2d interpolators of ion charge state densities
        """

        interpolators_abundance = interpolators2d_from_elementdensity(self.atomic_data, self.element, (self.r, self.z),
                                                                      self.n_element_2d,
                                                                      self.n_e_profile_2d, self.t_e_2d,
                                                                      tcx_donor=self.tcx_donor,
                                                                      tcx_donor_n=self.n_tcx_donor_2d,
                                                                      tcx_donor_charge=0)

        profiles = self.evaluate_interpolators(interpolators_abundance, (self.r, self.z))
        fraction_sum = self.sumup_fractions(profiles)

        self.assertTrue(np.allclose(fraction_sum, self.n_element_profile_2d, rtol=self.TOLERANCE))

    def test_balance_2d_interpolators_plasma_neutrality(self):
        """
        test calulation of 2d interpolators for ion charge state densities using plasma neutrality condition.
        """

        interpolators_abundance_1 = interpolators2d_from_elementdensity(self.atomic_data, self.element,
                                                                        (self.r, self.z),
                                                                        self.n_element_2d,
                                                                        self.n_e_profile_2d, self.t_e_2d)

        interpolators_abundance_2 = interpolators2d_from_elementdensity(self.atomic_data, self.element2,
                                                                        (self.r, self.z),
                                                                        self.n_element2_2d,
                                                                        self.n_e_profile_2d, self.t_e_2d)

        interpolators_abundance_3 = interpolators2d_match_plasma_neutrality(self.atomic_data, self.element,
                                                                            (self.r, self.z),
                                                                            [interpolators_abundance_1,
                                                                             interpolators_abundance_2],
                                                                            self.n_e_profile_2d, self.t_e_2d)

        profiles1 = self.evaluate_interpolators(interpolators_abundance_1, (self.r, self.z))
        profiles2 = self.evaluate_interpolators(interpolators_abundance_2, (self.r, self.z))
        profiles3 = self.evaluate_interpolators(interpolators_abundance_3, (self.r, self.z))

        total = self.sumup_electrons(profiles1)
        total += self.sumup_electrons(profiles2)
        total += self.sumup_electrons(profiles3)

        self.assertTrue(np.allclose(total, self.n_e_profile_2d, rtol=self.TOLERANCE))

    def test_balance_2d_interpolators_plasma_neutrality_tcx(self):
        """
        test calulation of 2d interpolators for ion charge state densities using plasma neutrality condition.
        """

        interpolators_abundance_1 = interpolators2d_from_elementdensity(self.atomic_data, self.element,
                                                                        (self.r, self.z),
                                                                        self.n_element_2d,
                                                                        self.n_e_profile_2d, self.t_e_2d,
                                                                        tcx_donor=self.tcx_donor,
                                                                        tcx_donor_n=self.n_tcx_donor_2d,
                                                                        tcx_donor_charge=0)

        interpolators_abundance_2 = interpolators2d_from_elementdensity(self.atomic_data, self.element2,
                                                                        (self.r, self.z),
                                                                        self.n_element2_2d,
                                                                        self.n_e_profile_2d, self.t_e_2d,
                                                                        tcx_donor=self.tcx_donor,
                                                                        tcx_donor_n=self.n_tcx_donor_2d,
                                                                        tcx_donor_charge=0)

        interpolators_abundance_3 = interpolators2d_match_plasma_neutrality(self.atomic_data, self.element,
                                                                            (self.r, self.z),
                                                                            [interpolators_abundance_1,
                                                                             interpolators_abundance_2],
                                                                            self.n_e_profile_2d, self.t_e_2d,
                                                                            tcx_donor=self.tcx_donor,
                                                                            tcx_donor_n=self.n_tcx_donor_2d,
                                                                            tcx_donor_charge=0)

        profiles1 = self.evaluate_interpolators(interpolators_abundance_1, (self.r, self.z))
        profiles2 = self.evaluate_interpolators(interpolators_abundance_2, (self.r, self.z))
        profiles3 = self.evaluate_interpolators(interpolators_abundance_3, (self.r, self.z))

        total = self.sumup_electrons(profiles1)
        total += self.sumup_electrons(profiles2)
        total += self.sumup_electrons(profiles3)

        self.assertTrue(np.allclose(total, self.n_e_profile_2d, rtol=self.TOLERANCE))

    def test_axisymmetric_mapper(self):
        "test axisymmetric mapping"

        # fractional abundance
        interpolators_fractional = interpolators2d_fractional(self.atomic_data, self.element, (self.r, self.z),
                                                              self.n_e_profile_2d, self.t_e_2d,
                                                              tcx_donor=self.tcx_donor,
                                                              tcx_donor_n=self.n_tcx_donor_profile_2d,
                                                              tcx_donor_charge=0)

        mappers = abundance_axisymmetric_mapper(interpolators_fractional)

        profile = self.evaluate_interpolators(mappers, self.r)

        total = self.sumup_fractions(profile)

        self.assertTrue(np.allclose(total, 1, rtol=self.TOLERANCE))

    def test_equilibrium_map3d_fractional(self):
        """
        test calculation of fractional abundance and application of map3d functionality of equilibrium
        """

        equilibrium = example_equilibrium()

        mapper = equilibrium_map3d_fractional(self.atomic_data, self.element, equilibrium, self.psin_1d,
                                              self.n_e_profile_1d, self.t_e_profile_1d, self.tcx_donor,
                                              self.n_tcx_donor_profile_1d, tcx_donor_charge=0)
        psin_1d = np.linspace(0, 0.99, 10)
        r = np.zeros_like(psin_1d)
        for index, value in enumerate(psin_1d):
            r[index] = equilibrium.psin_to_r(value)

        profile = self.evaluate_interpolators(mapper, r)

        total = self.sumup_fractions(profile)

        self.assertTrue(np.allclose(total, 1, rtol=self.TOLERANCE))

    def test_equilibrium_map3d_from_elementdensity(self):
        """
        test calculation of abundance and application of map3d functionality of equilibrium
        """

        equilibrium = example_equilibrium()

        mapper = equilibrium_map3d_from_elementdensity(self.atomic_data, self.element, equilibrium, self.psin_1d,
                                                       self.n_element_1d, self.n_e_1d,
                                                       self.t_e_1d, self.tcx_donor, self.n_tcx_donor_1d,
                                                       tcx_donor_charge=0)

        inlcfs = np.where(self.psin_1d < 1)[0]
        psin_1d = self.psin_1d[inlcfs]
        n = np.zeros_like(psin_1d)
        r = np.zeros_like(psin_1d)
        for index, value in enumerate(psin_1d):
            n[index] = self.n_element_1d(value)
            r[index] = equilibrium.psin_to_r(value)

        profile = self.evaluate_interpolators(mapper, r)
        total = self.sumup_fractions(profile)

        self.assertTrue(np.allclose(total, n, rtol=self.TOLERANCE))

    def test_equilibrium_map3d_plasma_neutrality(self):
        """
        test calculation of abundance and application of map3d functionality of equilibrium
        """

        equilibrium = example_equilibrium()

        interpolators_element1 = interpolators1d_from_elementdensity(self.atomic_data, self.element, self.psin_1d,
                                                                     self.n_element_1d,
                                                                     self.n_e_1d, self.t_e_1d,
                                                                     tcx_donor=self.tcx_donor,
                                                                     tcx_donor_n=self.n_tcx_donor_1d,
                                                                     tcx_donor_charge=0)

        interpolators_element2 = interpolators1d_from_elementdensity(self.atomic_data, self.element2, self.psin_1d,
                                                                     self.n_element2_1d,
                                                                     self.n_e_1d, self.t_e_1d,
                                                                     tcx_donor=self.tcx_donor,
                                                                     tcx_donor_n=self.n_tcx_donor_1d,
                                                                     tcx_donor_charge=0)

        mapper = equilibrium_map3d_match_plasma_neutrality(self.atomic_data, self.element_bulk, equilibrium, self.psin_1d,
                                                           [interpolators_element1, interpolators_element2],
                                                           self.n_e_1d, self.t_e_1d, self.tcx_donor,
                                                           self.n_tcx_donor_1d, 0)

        inlcfs = np.where(self.psin_1d < 1)[0]
        psin_1d = self.psin_1d[inlcfs]
        n = np.zeros_like(psin_1d)
        r = np.zeros_like(psin_1d)
        for index, value in enumerate(psin_1d):
            n[index] = self.n_e_1d(value)
            r[index] = equilibrium.psin_to_r(value)

        profile1 = self.evaluate_interpolators(interpolators_element1, psin_1d)
        profile2 = self.evaluate_interpolators(interpolators_element2, psin_1d)
        profile3 = self.evaluate_interpolators(mapper, r)

        total = self.sumup_electrons(profile1)
        total += self.sumup_electrons(profile2)
        total += self.sumup_electrons(profile3)

        self.assertTrue(np.allclose(total, n, rtol=self.TOLERANCE))


if __name__ == "__main__":
    unittest.main()




import numpy as np
import matplotlib.pyplot as plt

from raysect.core import Point2D, translate, Vector3D, rotate_basis
from raysect.primitive import Cylinder
from raysect.optical import World
from raysect.optical.observer import PinholeCamera, PowerPipeline2D
from raysect.optical.material import VolumeTransform

from cherab.core.math import sample2d, AxisymmetricMapper
from cherab.tools.emitters import RadiationFunction


#############################
# define radiation function #


PLASMA_AXIS = Point2D(1.5, 0)
LCFS_RADIUS = 1
RING_RADIUS = 0.5

RADIATION_PEAK = 1e6
CENTRE_PEAK_WIDTH = 0.05
RING_WIDTH = 0.025

# distance of wall from LCFS
WALL_LCFS_OFFSET = 0.1

CYLINDER_RADIUS = PLASMA_AXIS.x + LCFS_RADIUS + WALL_LCFS_OFFSET * 1.1
CYLINDER_HEIGHT = (LCFS_RADIUS + WALL_LCFS_OFFSET) * 2


def rad_function(r, z):

    sample_point = Point2D(r, z)
    direction = PLASMA_AXIS.vector_to(sample_point)
    bearing = np.arctan2(direction.y, direction.x)

    # calculate radius of coordinate from magnetic axis
    radius_from_axis = direction.length
    closest_ring_point = PLASMA_AXIS + (direction.normalise() * 0.5)
    radius_from_ring = sample_point.distance_to(closest_ring_point)

    # evaluate pedestal-> core function
    if radius_from_axis <= LCFS_RADIUS:

        central_radiatior = RADIATION_PEAK * np.exp(-(radius_from_axis ** 2) / CENTRE_PEAK_WIDTH)

        ring_radiator = RADIATION_PEAK * np.cos(bearing) * np.exp(-(radius_from_ring ** 2) / RING_WIDTH)
        ring_radiator = max(0, ring_radiator)

        return central_radiatior + ring_radiator
    else:
        return 0


#################################
# add radiation source to world #

world = World()

rad_function_3d = AxisymmetricMapper(rad_function)
# We shift the cylinder containing the emission function relative to the world,
# so need to apply the opposite shift to the material to ensure the radiation
# function is evaluated in the correct coordinate system.
shift = translate(0, 0, -1)
radiation_emitter = VolumeTransform(RadiationFunction(rad_function_3d), shift.inverse())
geom = Cylinder(CYLINDER_RADIUS, CYLINDER_HEIGHT,
                transform=shift, parent=world, material=radiation_emitter)


######################
# visualise emission #

# run some plots to check the distribution functions and emission profile are as expected
r, z, t_samples = sample2d(rad_function, (0, 4, 200), (-2, 2, 200))
plt.imshow(np.transpose(np.squeeze(t_samples)), extent=[0, 3, -1.5, 1.5])
plt.colorbar()
plt.axis('equal')
plt.xlabel('r axis')
plt.ylabel('z axis')
plt.title("Radiation profile in r-z plane")


camera = PinholeCamera((256, 256), pipelines=[PowerPipeline2D()], parent=world)
camera.transform = translate(-3.5, -1.5, 0)*rotate_basis(Vector3D(1, 0, 0), Vector3D(0, 0, 1))
camera.pixel_samples = 1

plt.ion()
camera.observe()
plt.ioff()
plt.show()



# Copyright 2016-2022 Euratom
# Copyright 2016-2022 United Kingdom Atomic Energy Authority
# Copyright 2016-2022 Centro de Investigaciones Energéticas, Medioambientales y Tecnológicas
#
# Licensed under the EUPL, Version 1.1 or – as soon they will be approved by the
# European Commission - subsequent versions of the EUPL (the "Licence");
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at:
#
# https://joinup.ec.europa.eu/software/page/eupl5
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the Licence is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.
#
# See the Licence for the specific language governing permissions and limitations
# under the Licence.

cdef:

    # sourced c standard maths library
    double RECIP_2_PI = 1 / (2 * M_PI)
    double RECIP_4_PI = 1 / (4 * M_PI)
    double DEGREES_TO_RADIANS = M_PI / 180
    double RADIANS_TO_DEGREES = 180 / M_PI

    # sourced from NIST, CODATA 2018: https://physics.nist.gov/cuu/Constants/Table/allascii.txt
    double ATOMIC_MASS = 1.66053906660e-27
    double ELEMENTARY_CHARGE = 1.602176634e-19
    double SPEED_OF_LIGHT = 299792458.0
    double PLANCK_CONSTANT = 6.62607015e-34
    double HC_EV_NM = 1239.8419738620933  # (Planck constant in eV s) x (speed of light in nm/s)
    double ELECTRON_CLASSICAL_RADIUS = 2.8179403262e-15
    double ELECTRON_REST_MASS = 9.1093837015e-31
    double RYDBERG_CONSTANT_EV = 13.605693122994
    double VACUUM_PERMITTIVITY = 8.8541878128e-12
    double BOHR_MAGNETON = 5.78838180123e-5  # in eV/T



# Copyright 2016-2018 Euratom
# Copyright 2016-2018 United Kingdom Atomic Energy Authority
# Copyright 2016-2018 Centro de Investigaciones Energéticas, Medioambientales y Tecnológicas
#
# Licensed under the EUPL, Version 1.1 or – as soon they will be approved by the
# European Commission - subsequent versions of the EUPL (the "Licence");
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at:
#
# https://joinup.ec.europa.eu/software/page/eupl5
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the Licence is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.
#
# See the Licence for the specific language governing permissions and limitations
# under the Licence.

"""
This script has been used to calculate the reference data for the 3D interpolators tests.
To get the reference data for the 3D linear interpolators, uncomment the linear
part (l.210) and comment the cubic part (l.287). To get the reference data for
the 3D cubic interpolator, do the contrary.
Reference interpolated data is stored in 'interp_data'.
Reference extrapolated data with nearest method is stored in 'extrap_data_nea'
Reference extrapolated data with linear method is stored in 'extrap_data_lin'
Reference extrapolated data with quadratic method is stored in 'extrap_data_qua'
The function 'write_in_file' has been made to write extrapolation data in a nice way.
"""

import matplotlib.pyplot as plt
import numpy as np
from scipy.linalg import inv
from numpy.linalg import tensorsolve

import sys
# sys.path.append('/home/cbert/code/Raysect')
# from cherab.math.mapping.interpolators import interpolators3d

np.set_printoptions(12, 30000, linewidth=340, formatter={'float': lambda x: format(x, ' 0.12e')})  # 15 values in a line

NSX, NSY, NSZ = 15, 10, 15

x = np.linspace(0, 1, 10)
y = np.linspace(0, 2, 5)
z = np.linspace(-1, 1, 5)
x2 = x*x
x3 = x2*x
y2 = y*y
y3 = y2*y
z2 = z*z
z3 = z2*z
xsamples = np.linspace(0, 1, NSX)
ysamples = np.linspace(0, 2, NSY)
zsamples = np.linspace(-1, 1, NSZ)
xsamples_ex = np.array([0 - 0.08, 0 - 0.05, 0 - 0.02] + list(xsamples) + [1 + 0.02, 1 + 0.05, 1 + 0.08], dtype=np.float64)
ysamples_ex = np.array([0 - 0.16, 0 - 0.10, 0 - 0.04] + list(ysamples) + [2 + 0.04, 2 + 0.10, 2 + 0.16], dtype=np.float64)
zsamples_ex = np.array([-1 - 0.16, -1 - 0.10, -1 - 0.04] + list(zsamples) + [1 + 0.04, 1 + 0.10, 1 + 0.16], dtype=np.float64)
extrapol_xdomains = [(0, 3), (3, NSX + 3), (NSX + 3, NSX + 6)]
extrapol_ydomains = [(0, 3), (3, NSY + 3), (NSY + 3, NSY + 6)]
extrapol_zdomains = [(0, 3), (3, NSZ + 3), (NSZ + 3, NSZ + 6)]

deltax = x[2] - x[0]
deltay = y[2] - y[0]
deltaz = z[2] - z[0]


def point_to_indices(a, b, c):
    ia = 0
    ib = 0
    ic = 0
    if a < x[0] or a > x[-1] or b < y[0] or b > y[-1] or c < z[0] or c > z[-1]:
        raise ValueError("point outside interpolation area")
    while not (x[ia] <= a <= x[ia+1]):
        ia += 1
    while not (y[ib] <= b <= y[ib+1]):
        ib += 1
    while not (z[ic] <= c <= z[ic+1]):
        ic += 1

    return ia, ib, ic

interp_data = np.zeros((NSX, NSY, NSZ), dtype=np.float64)

extrap_data_nea = [[[np.zeros((3, 3, 3), dtype=np.float64),
                     np.zeros((3, 3, NSZ), dtype=np.float64),
                     np.zeros((3, 3, 3), dtype=np.float64)],
                    [np.zeros((3, NSY, 3), dtype=np.float64),
                     np.zeros((3, NSY, NSZ), dtype=np.float64),
                     np.zeros((3, NSY, 3), dtype=np.float64)],
                    [np.zeros((3, 3, 3), dtype=np.float64),
                     np.zeros((3, 3, NSZ), dtype=np.float64),
                     np.zeros((3, 3, 3), dtype=np.float64)]],
                   [[np.zeros((NSX, 3, 3), dtype=np.float64),
                     np.zeros((NSX, 3, NSZ), dtype=np.float64),
                     np.zeros((NSX, 3, 3), dtype=np.float64)],
                    [np.zeros((NSX, NSY, 3), dtype=np.float64),
                     None,
                     np.zeros((NSX, NSY, 3), dtype=np.float64)],
                    [np.zeros((NSX, 3, 3), dtype=np.float64),
                     np.zeros((NSX, 3, NSZ), dtype=np.float64),
                     np.zeros((NSX, 3, 3), dtype=np.float64)]],
                   [[np.zeros((3, 3, 3), dtype=np.float64),
                     np.zeros((3, 3, NSZ), dtype=np.float64),
                     np.zeros((3, 3, 3), dtype=np.float64)],
                    [np.zeros((3, NSY, 3), dtype=np.float64),
                     np.zeros((3, NSY, NSZ), dtype=np.float64),
                     np.zeros((3, NSY, 3), dtype=np.float64)],
                    [np.zeros((3, 3, 3), dtype=np.float64),
                     np.zeros((3, 3, NSZ), dtype=np.float64),
                     np.zeros((3, 3, 3), dtype=np.float64)]]]

extrap_data_lin = [[[np.zeros((3, 3, 3), dtype=np.float64),
                     np.zeros((3, 3, NSZ), dtype=np.float64),
                     np.zeros((3, 3, 3), dtype=np.float64)],
                    [np.zeros((3, NSY, 3), dtype=np.float64),
                     np.zeros((3, NSY, NSZ), dtype=np.float64),
                     np.zeros((3, NSY, 3), dtype=np.float64)],
                    [np.zeros((3, 3, 3), dtype=np.float64),
                     np.zeros((3, 3, NSZ), dtype=np.float64),
                     np.zeros((3, 3, 3), dtype=np.float64)]],
                   [[np.zeros((NSX, 3, 3), dtype=np.float64),
                     np.zeros((NSX, 3, NSZ), dtype=np.float64),
                     np.zeros((NSX, 3, 3), dtype=np.float64)],
                    [np.zeros((NSX, NSY, 3), dtype=np.float64),
                     None,
                     np.zeros((NSX, NSY, 3), dtype=np.float64)],
                    [np.zeros((NSX, 3, 3), dtype=np.float64),
                     np.zeros((NSX, 3, NSZ), dtype=np.float64),
                     np.zeros((NSX, 3, 3), dtype=np.float64)]],
                   [[np.zeros((3, 3, 3), dtype=np.float64),
                     np.zeros((3, 3, NSZ), dtype=np.float64),
                     np.zeros((3, 3, 3), dtype=np.float64)],
                    [np.zeros((3, NSY, 3), dtype=np.float64),
                     np.zeros((3, NSY, NSZ), dtype=np.float64),
                     np.zeros((3, NSY, 3), dtype=np.float64)],
                    [np.zeros((3, 3, 3), dtype=np.float64),
                     np.zeros((3, 3, NSZ), dtype=np.float64),
                     np.zeros((3, 3, 3), dtype=np.float64)]]]

extrap_data_qua = [[[np.zeros((3, 3, 3), dtype=np.float64),
                     np.zeros((3, 3, NSZ), dtype=np.float64),
                     np.zeros((3, 3, 3), dtype=np.float64)],
                    [np.zeros((3, NSY, 3), dtype=np.float64),
                     np.zeros((3, NSY, NSZ), dtype=np.float64),
                     np.zeros((3, NSY, 3), dtype=np.float64)],
                    [np.zeros((3, 3, 3), dtype=np.float64),
                     np.zeros((3, 3, NSZ), dtype=np.float64),
                     np.zeros((3, 3, 3), dtype=np.float64)]],
                   [[np.zeros((NSX, 3, 3), dtype=np.float64),
                     np.zeros((NSX, 3, NSZ), dtype=np.float64),
                     np.zeros((NSX, 3, 3), dtype=np.float64)],
                    [np.zeros((NSX, NSY, 3), dtype=np.float64),
                     None,
                     np.zeros((NSX, NSY, 3), dtype=np.float64)],
                    [np.zeros((NSX, 3, 3), dtype=np.float64),
                     np.zeros((NSX, 3, NSZ), dtype=np.float64),
                     np.zeros((NSX, 3, 3), dtype=np.float64)]],
                   [[np.zeros((3, 3, 3), dtype=np.float64),
                     np.zeros((3, 3, NSZ), dtype=np.float64),
                     np.zeros((3, 3, 3), dtype=np.float64)],
                    [np.zeros((3, NSY, 3), dtype=np.float64),
                     np.zeros((3, NSY, NSZ), dtype=np.float64),
                     np.zeros((3, NSY, 3), dtype=np.float64)],
                    [np.zeros((3, 3, 3), dtype=np.float64),
                     np.zeros((3, 3, NSZ), dtype=np.float64),
                     np.zeros((3, 3, 3), dtype=np.float64)]]]

def f(a, b, c):
    return (a-b)*b*np.cos(10*a)*np.exp(c) + c

data = np.zeros((10, 5, 5), dtype=np.float64)
for i in range(10):
    for j in range(5):
        for k in range(5):
            data[i, j, k] = f(x[i], y[j], z[k])

def write_in_file(data, filename, prefix='    '):

    to_print = ''
    positions = ['inf', 'mid', 'sup']
    tab = '    '
    p = 0

    to_print += prefix + '[\n'
    for x_arrays in data:
        px = p // 9
        to_print += prefix + tab + '[  # x {}\n'.format(positions[px])
        for y_arrays in x_arrays:
            py = (p // 3) % 3
            to_print += prefix + 2 * tab + '[  # y {}\n'.format(positions[py])
            for z_arrays in y_arrays:
                pz = p % 3
                if z_arrays is not None:
                    to_print += prefix + 3 * tab + 'np.array([  # z {}\n'.format(positions[pz])
                    for col in z_arrays:
                        first = True
                        for raw in col:
                            to_print += prefix + tab + '[' if not first else prefix + tab[:-1] + '[['
                            first = False
                            for elt in raw:
                                to_print += '{: 0.12e}, '.format(elt)
                            to_print = to_print[:-2]  # drop last coma
                            to_print += '],\n'
                        to_print = to_print[:-2] + '],\n'  # add a bracket
                    to_print = to_print[:-2] + '\n'  # drop last coma
                    to_print += prefix + 3 * tab + '], dtype=np.float64)'
                    to_print += ',\n' if pz < 2 else '\n'
                else:
                    to_print += '\n' + prefix + 3* tab + 'None,  # z mid (interpolation area)\n\n'
                p += 1
            to_print += prefix + 2 * tab + ']'
            to_print += ',\n' if py < 2 else '\n'
        to_print += prefix + tab + ']'
        to_print += ',\n' if px < 2 else '\n'
    to_print += prefix + ']\n'

    with open(filename, 'w') as file:
        orig_stdout = sys.stdout
        sys.stdout = file

        print(to_print)

        sys.stdout = orig_stdout

# ##### linear ##### #

def lin(d1, d2, delta, pos):
    return d1 + pos * (d2 - d1) / delta

def P(i, j, k):

    return lambda a, b, c: lin(lin(lin(data[i, j, k], data[i+1, j, k], deltax/2, a - x[i]),
                                   lin(data[i, j+1, k], data[i+1, j+1, k], deltax/2, a - x[i]),
                                   deltay/2, b - y[j]),
                               lin(lin(data[i, j, k+1], data[i+1, j, k+1], deltax/2, a - x[i]),
                                   lin(data[i, j+1, k+1], data[i+1, j+1, k+1], deltax/2, a - x[i]),
                                   deltay/2, b - y[j]),
                               deltaz/2, c - z[k])

def interp_func(a, b, c):
    return P(*point_to_indices(a, b, c))(a, b, c)

for i in range(NSX):
    for j in range(NSY):
        for k in range(NSZ):
            interp_data[i, j, k] = interp_func(xsamples[i], ysamples[j], zsamples[k])

def e0(a, b, c):
    a_n = max(min(a, x.max()), x.min())
    b_n = max(min(b, y.max()), y.min())
    c_n = max(min(c, z.max()), z.min())
    return interp_func(a_n, b_n, c_n)

def e1(a, b, c):
    a_n = max(min(a, x.max()), x.min())
    b_n = max(min(b, y.max()), y.min())
    c_n = max(min(c, z.max()), z.min())
    k = max(abs(a - a_n), abs(b - b_n), abs(c - c_n))
    a_ref = a_n - (a - a_n) / k
    b_ref = b_n - (b - b_n) / k
    c_ref = c_n - (c - c_n) / k

    func = P(*point_to_indices(a_n, b_n, c_n))

    da = (func(a_n, b_n, c_n) - func(a_ref, b_n, c_n)) / (a_n - a_ref) if a_n != a_ref else 0.
    db = (func(a_n, b_n, c_n) - func(a_n, b_ref, c_n)) / (b_n - b_ref) if b_n != b_ref else 0.
    dc = (func(a_n, b_n, c_n) - func(a_n, b_n, c_ref)) / (c_n - c_ref) if c_n != c_ref else 0.
    dadb = (func(a_n, b_n, c_n) - func(a_n, b_ref, c_n) - func(a_ref, b_n, c_n) + func(a_ref, b_ref, c_n)) / (a_n - a_ref) / (b_n - b_ref) if a_n != a_ref and b_n != b_ref else 0.
    dbdc = (func(a_n, b_n, c_n) - func(a_n, b_ref, c_n) - func(a_n, b_n, c_ref) + func(a_n, b_ref, c_ref)) / (c_n - c_ref) / (b_n - b_ref) if c_n != c_ref and b_n != b_ref else 0.
    dcda = (func(a_n, b_n, c_n) - func(a_n, b_n, c_ref) - func(a_ref, b_n, c_n) + func(a_ref, b_n, c_ref)) / (a_n - a_ref) / (c_n - c_ref) if a_n != a_ref and c_n != c_ref else 0.
    dadbdc = (func(a_n, b_n, c_n) - func(a_n, b_ref, c_n) - func(a_ref, b_n, c_n) + func(a_ref, b_ref, c_n) - func(a_n, b_n, c_ref) + func(a_n, b_ref, c_ref) + func(a_ref, b_n, c_ref) - func(a_ref, b_ref, c_ref)) / (a_n - a_ref) / (b_n - b_ref) / (c_n - c_ref) if a_n != a_ref and b_n != b_ref and c_n != c_ref else 0.

    return func(a_n, b_n, c_n) + \
           (a - a_n) * da + \
           (b - b_n) * db + \
           (c - c_n) * dc + \
           (a - a_n) * (b - b_n) * dadb + \
           (b - b_n) * (c - c_n) * dbdc + \
           (c - c_n) * (a - a_n) * dcda + \
           (a - a_n) * (b - b_n) * (c - c_n) * dadbdc

whole_calc_data_nea = np.zeros((NSX+6, NSY+6, NSZ+6), dtype=np.float64)
whole_calc_data_nea[3:NSX+3, 3:NSY+3, 3:NSZ+3] = interp_data
whole_calc_data_lin = np.zeros((NSX+6, NSY+6, NSZ+6), dtype=np.float64)
whole_calc_data_lin[3:NSX+3, 3:NSY+3, 3:NSZ+3] = interp_data

for nx in range(3):
    for ny in range(3):
        for nz in range(3):
            if nx != 1 or ny != 1 or nz != 1:
                mini, maxi = extrapol_xdomains[nx]
                minj, maxj = extrapol_ydomains[ny]
                mink, maxk = extrapol_zdomains[nz]
                for iex in range(mini, maxi):
                    for jex in range(minj, maxj):
                        for kex in range(mink, maxk):
                            extrap_data_nea[nx][ny][nz][iex - mini, jex - minj, kex - mink] = e0(xsamples_ex[iex], ysamples_ex[jex], zsamples_ex[kex])
                            extrap_data_lin[nx][ny][nz][iex - mini, jex - minj, kex - mink] = e1(xsamples_ex[iex], ysamples_ex[jex], zsamples_ex[kex])
                whole_calc_data_nea[mini:maxi, minj:maxj, mink:maxk] = extrap_data_nea[nx][ny][nz]
                whole_calc_data_lin[mini:maxi, minj:maxj, mink:maxk] = extrap_data_lin[nx][ny][nz]

# ##### cubic ##### #

# def elt_mat(i, j, k):
#     output = np.zeros((64, 64), dtype=np.float64)
#
#     for ic in range(4):
#         for jc in range(4):
#             for kc in range(4):
#                 l = 0
#                 for u in [i, i+1]:
#                     for v in [j, j+1]:
#                         for w in [k, k+1]:
#                             for dx in [0, 1]:
#                                 for dy in [0, 1]:
#                                     for dz in [0, 1]:
#                                         output[l, ic + 4*jc + 16*kc] = x[u]**(abs(ic-dx))*(ic**dx) * y[v]**(abs(jc-dy))*(jc**dy) * z[w]**(abs(kc-dz))*(kc**dz)
#                                         l += 1
#     return output
#
# def d(i, j, k):
#     ieff = max(min(i, 9), 0)
#     jeff = max(min(j, 4), 0)
#     keff = max(min(k, 4), 0)
#     return data[ieff, jeff, keff]
#
# def xf(i):
#     ieff = max(min(i, 9), 0)
#     return x[ieff]
#
# def yf(j):
#     jeff = max(min(j, 4), 0)
#     return y[jeff]
#
# def zf(k):
#     keff = max(min(k, 4), 0)
#     return z[keff]
#
# def elt_vec(i, j, k):
#
#     output = np.zeros((64,), dtype=np.float64)
#
#     l = 0
#     for u in [i, i+1]:
#         for v in [j, j+1]:
#             for w in [k, k+1]:
#                 for dx in [0, 1]:
#                     for dy in [0, 1]:
#                         for dz in [0, 1]:
#
#                             value = 0.
#                             for a in {-dx, dx}:
#                                 for b in {-dy, dy}:
#                                     for c in {-dz, dz}:
#                                         value += (a+0.5)/abs(a+0.5)*(b+0.5)/abs(b+0.5)*(c+0.5)/abs(c+0.5)*d(u+a, v+b, w+c)
#                             value /= (xf(u+1) - xf(u-1))**dx * (yf(v+1) - yf(v-1))**dy * (zf(w+1) - zf(w-1))**dz
#                             output[l] = value
#                             l += 1
#
#     return output
#
# coeffs = []
#
# for i in range(9):
#     coeffs_i = []
#     for j in range(4):
#         coeffs_ij = []
#         for k in range(4):
#             coeffs_ijk = tensorsolve(elt_mat(i, j, k), elt_vec(i, j, k))
#             coeffs_ij.append(coeffs_ijk)
#         coeffs_i.append(coeffs_ij)
#     coeffs.append(coeffs_i)
#
# def P(i, j, k):
#     if 0 <= i <= 8 and 0 <= j <= 3 and 0 <= k <= 3:
#
#         def output(a, b, c):
#             result = 0.
#             for ic in range(4):
#                 for jc in range(4):
#                     for kc in range(4):
#                         result += coeffs[i][j][k][ic + 4*jc + 16*kc] * a**ic * b**jc * c**kc
#             return result
#
#         return output
#
#     else:
#         raise ValueError("i, j and k must be in the good range")
#
# def dP_dx(i, j, k):
#     if 0 <= i <= 8 and 0 <= j <= 3 and 0 <= k <= 3:
#
#         def output(a, b, c):
#             result = 0.
#             for ic in range(4):
#                 for jc in range(4):
#                     for kc in range(4):
#                         result += coeffs[i][j][k][ic + 4*jc + 16*kc] * ic * a**(abs(ic-1)) * b**jc * c**kc
#             return result
#
#         return output
#
#     else:
#         raise ValueError("i, j and k must be in the good range")
#
# def d2P_dx2(i, j, k):
#     if 0 <= i <= 8 and 0 <= j <= 3 and 0 <= k <= 3:
#
#         def output(a, b, c):
#             result = 0.
#             for ic in range(4):
#                 for jc in range(4):
#                     for kc in range(4):
#                         result += coeffs[i][j][k][ic + 4*jc + 16*kc] * ic*(ic-1) * a**(abs(ic-2)) * b**jc * c**kc
#             return result
#
#         return output
#
#     else:
#         raise ValueError("i, j and k must be in the good range")
#
# def dP_dy(i, j, k):
#     if 0 <= i <= 8 and 0 <= j <= 3 and 0 <= k <= 3:
#
#         def output(a, b, c):
#             result = 0.
#             for ic in range(4):
#                 for jc in range(4):
#                     for kc in range(4):
#                         result += coeffs[i][j][k][ic + 4*jc + 16*kc] * jc * a**ic * b**(abs(jc-1)) * c**kc
#             return result
#
#         return output
#
#     else:
#         raise ValueError("i, j and k must be in the good range")
#
# def d2P_dy2(i, j, k):
#     if 0 <= i <= 8 and 0 <= j <= 3 and 0 <= k <= 3:
#
#         def output(a, b, c):
#             result = 0.
#             for ic in range(4):
#                 for jc in range(4):
#                     for kc in range(4):
#                         result += coeffs[i][j][k][ic + 4*jc + 16*kc] * jc*(jc-1) * a**ic * b**(abs(jc-2)) * c**kc
#             return result
#
#         return output
#
#     else:
#         raise ValueError("i, j and k must be in the good range")
#
# def dP_dz(i, j, k):
#     if 0 <= i <= 8 and 0 <= j <= 3 and 0 <= k <= 3:
#
#         def output(a, b, c):
#             result = 0.
#             for ic in range(4):
#                 for jc in range(4):
#                     for kc in range(4):
#                         result += coeffs[i][j][k][ic + 4*jc + 16*kc] * kc * a**ic * b**jc * c**(abs(kc-1))
#             return result
#
#         return output
#
#     else:
#         raise ValueError("i, j and k must be in the good range")
#
# def d2P_dz2(i, j, k):
#     if 0 <= i <= 8 and 0 <= j <= 3 and 0 <= k <= 3:
#
#         def output(a, b, c):
#             result = 0.
#             for ic in range(4):
#                 for jc in range(4):
#                     for kc in range(4):
#                         result += coeffs[i][j][k][ic + 4*jc + 16*kc] * kc*(kc-1) * a**ic * b**(jc) * c**(abs(kc-2))
#             return result
#
#         return output
#
#     else:
#         raise ValueError("i, j and k must be in the good range")
#
# def d2P_dxdy(i, j, k):
#     if 0 <= i <= 8 and 0 <= j <= 3 and 0 <= k <= 3:
#
#         def output(a, b, c):
#             result = 0.
#             for ic in range(4):
#                 for jc in range(4):
#                     for kc in range(4):
#                         result += coeffs[i][j][k][ic + 4*jc + 16*kc] * ic*jc * a**(abs(ic-1)) * b**(abs(jc-1)) * c**kc
#             return result
#
#         return output
#
#     else:
#         raise ValueError("i, j and k must be in the good range")
#
# def d2P_dydz(i, j, k):
#     if 0 <= i <= 8 and 0 <= j <= 3 and 0 <= k <= 3:
#
#         def output(a, b, c):
#             result = 0.
#             for ic in range(4):
#                 for jc in range(4):
#                     for kc in range(4):
#                         result += coeffs[i][j][k][ic + 4*jc + 16*kc] * kc*jc * a**ic * b**(abs(jc-1)) * c**(abs(kc-1))
#             return result
#
#         return output
#
#     else:
#         raise ValueError("i, j and k must be in the good range")
#
# def d2P_dzdx(i, j, k):
#     if 0 <= i <= 8 and 0 <= j <= 3 and 0 <= k <= 3:
#
#         def output(a, b, c):
#             result = 0.
#             for ic in range(4):
#                 for jc in range(4):
#                     for kc in range(4):
#                         result += coeffs[i][j][k][ic + 4*jc + 16*kc] * kc*ic * a**(abs(ic-1)) * b**jc * c**(abs(kc-1))
#             return result
#
#         return output
#
#     else:
#         raise ValueError("i, j and k must be in the good range")
#
#
# def interp_func(a, b, c):
#     return P(*point_to_indices(a, b, c))(a, b, c)
#
# for i in range(NSX):
#     for j in range(NSY):
#         for k in range(NSZ):
#             interp_data[i, j, k] = interp_func(xsamples[i], ysamples[j], zsamples[k])
#
# # compare_func = interpolators3d.Interpolate3DLinear(x, y, z, data)
# # compare_data = np.zeros((NSX, NSY, NSZ), dtype=np.float64)
# # for i in range(NSX):
# #     for j in range(NSY):
# #         for k in range(NSZ):
# #             compare_data[i, j, k] = compare_func(xsamples[i], ysamples[j], zsamples[k])
#
#
# def e0(a, b, c):
#     a_n = max(min(a, x.max()), x.min())
#     b_n = max(min(b, y.max()), y.min())
#     c_n = max(min(c, z.max()), z.min())
#     return interp_func(a_n, b_n, c_n)
#
# def e1(a, b, c):
#     a_n = max(min(a, x.max()), x.min())
#     b_n = max(min(b, y.max()), y.min())
#     c_n = max(min(c, z.max()), z.min())
#     i, j, k = point_to_indices(a_n, b_n, c_n)
#
#     return P(i, j, k)(a_n, b_n, c_n) + \
#            (a - a_n) * dP_dx(i, j, k)(a_n, b_n, c_n) + \
#            (b - b_n) * dP_dy(i, j, k)(a_n, b_n, c_n) + \
#            (c - c_n) * dP_dz(i, j, k)(a_n, b_n, c_n)
#
# def e2(a, b, c):
#     a_n = max(min(a, x.max()), x.min())
#     b_n = max(min(b, y.max()), y.min())
#     c_n = max(min(c, z.max()), z.min())
#     i, j, k = point_to_indices(a_n, b_n, c_n)
#
#     return e1_cub(a, b, c) + \
#            (a - a_n) * (b - b_n) * d2P_dxdy(i, j, k)(a_n, b_n, c_n) + \
#            (b - b_n) * (c - c_n) * d2P_dydz(i, j, k)(a_n, b_n, c_n) + \
#            (c - c_n) * (a - a_n) * d2P_dzdx(i, j, k)(a_n, b_n, c_n) + \
#            0.5 * (a - a_n)**2 * d2P_dx2(i, j, k)(a_n, b_n, c_n) + \
#            0.5 * (b - b_n)**2 * d2P_dy2(i, j, k)(a_n, b_n, c_n) + \
#            0.5 * (c - c_n)**2 * d2P_dz2(i, j, k)(a_n, b_n, c_n)
#
#
#
# whole_calc_data_nea = np.zeros((NSX+6, NSY+6, NSZ+6), dtype=np.float64)
# whole_calc_data_nea[3:NSX+3, 3:NSY+3, 3:NSZ+3] = interp_data
# whole_calc_data_lin = np.zeros((NSX+6, NSY+6, NSZ+6), dtype=np.float64)
# whole_calc_data_lin[3:NSX+3, 3:NSY+3, 3:NSZ+3] = interp_data
# whole_calc_data_qua = np.zeros((NSX+6, NSY+6, NSZ+6), dtype=np.float64)
# whole_calc_data_qua[3:NSX+3, 3:NSY+3, 3:NSZ+3] = interp_data
#
# for nx in range(3):
#     for ny in range(3):
#         for nz in range(3):
#             if nx != 1 or ny != 1 or nz != 1:
#                 mini, maxi = extrapol_xdomains[nx]
#                 minj, maxj = extrapol_ydomains[ny]
#                 mink, maxk = extrapol_zdomains[nz]
#                 for iex in range(mini, maxi):
#                     for jex in range(minj, maxj):
#                         for kex in range(mink, maxk):
#                             extrap_data_nea[nx][ny][nz][iex - mini, jex - minj, kex - mink] = e0(xsamples_ex[iex], ysamples_ex[jex], zsamples_ex[kex])
#                             extrap_data_lin[nx][ny][nz][iex - mini, jex - minj, kex - mink] = e1(xsamples_ex[iex], ysamples_ex[jex], zsamples_ex[kex])
#                             extrap_data_qua[nx][ny][nz][iex - mini, jex - minj, kex - mink] = e2(xsamples_ex[iex], ysamples_ex[jex], zsamples_ex[kex])
#                 whole_calc_data_nea[mini:maxi, minj:maxj, mink:maxk] = extrap_data_nea[nx][ny][nz]
#                 whole_calc_data_lin[mini:maxi, minj:maxj, mink:maxk] = extrap_data_lin[nx][ny][nz]
#                 whole_calc_data_qua[mini:maxi, minj:maxj, mink:maxk] = extrap_data_qua[nx][ny][nz]
#
#
# # def compare():
# #     for k in range(NSZ):
# #         plt.figure()
# #         plt.contourf(compare_data[:, :, k] - interp_data[:, :, k])
# #         plt.colorbar()
# #         plt.show()
#



# Copyright 2016-2022 Euratom
# Copyright 2016-2022 United Kingdom Atomic Energy Authority
# Copyright 2016-2022 Centro de Investigaciones Energéticas, Medioambientales y Tecnológicas
#
# Licensed under the EUPL, Version 1.1 or – as soon they will be approved by the
# European Commission - subsequent versions of the EUPL (the "Licence");
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at:
#
# https://joinup.ec.europa.eu/software/page/eupl5
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the Licence is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.
#
# See the Licence for the specific language governing permissions and limitations
# under the Licence.

from raysect.core.math.function.float import Exp1D, Arg1D
from cherab.core.math.integrators import GaussianQuadrature
from math import sqrt, pi
from scipy.special import erf
import unittest


class TestGaussianQuadrature(unittest.TestCase):
    """Gaussian quadrature integrator tests."""

    def test_properties(self):
        """Test property assignment."""
        min_order = 3
        max_order = 30
        reltol = 1.e-6
        quadrature = GaussianQuadrature(integrand=Arg1D, relative_tolerance=reltol, max_order=max_order, min_order=min_order)

        self.assertEqual(quadrature.relative_tolerance, reltol)
        self.assertEqual(quadrature.max_order, max_order)
        self.assertEqual(quadrature.min_order, min_order)
        self.assertEqual(quadrature.integrand, Arg1D)

        min_order = 0
        max_order = 2  # < min_order
        reltol = -1

        with self.assertRaises(ValueError):
            quadrature.max_order = max_order

        with self.assertRaises(ValueError):
            quadrature.min_order = min_order

        with self.assertRaises(ValueError):
            quadrature.relative_tolerance = reltol

        min_order = 1
        max_order = 20
        reltol = 1.e-5

        quadrature.relative_tolerance = reltol
        quadrature.min_order = min_order
        quadrature.max_order = max_order
        quadrature.integrand = Exp1D

        self.assertEqual(quadrature.relative_tolerance, reltol)
        self.assertEqual(quadrature.min_order, min_order)
        self.assertEqual(quadrature.max_order, max_order)
        self.assertEqual(quadrature.integrand, Exp1D)

    def test_integrate(self):
        """Test integration."""
        quadrature = GaussianQuadrature(relative_tolerance=1.e-8)
        a = -0.5
        b = 3.
        quadrature.integrand = (2 / sqrt(pi)) * Exp1D(- Arg1D() * Arg1D())
        exact_integral = erf(b) - erf(a)

        self.assertAlmostEqual(quadrature(a, b), exact_integral, places=8)


if __name__ == '__main__':
    unittest.main()



# Copyright 2016-2018 Euratom
# Copyright 2016-2018 United Kingdom Atomic Energy Authority
# Copyright 2016-2018 Centro de Investigaciones Energéticas, Medioambientales y Tecnológicas
#
# Licensed under the EUPL, Version 1.1 or – as soon they will be approved by the
# European Commission - subsequent versions of the EUPL (the "Licence");
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at:
#
# https://joinup.ec.europa.eu/software/page/eupl5
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the Licence is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.
#
# See the Licence for the specific language governing permissions and limitations
# under the Licence.

"""
EFIT Equilibrium Object Demonstration
-------------------------------------

This file will demonstrate how to:

 * access equilibrium attributes
 * map functions onto flux surfaces (2d and 3d)

"""


import matplotlib.pyplot as plt
from raysect.core.math.function.float import Interpolator1DArray
from cherab.core.math import sample2d, Slice3D
from cherab.tools.equilibrium import example_equilibrium, plot_equilibrium


# populate and return an example equilibrium object.
print('Populating example equilibrium object...')
equilibrium = example_equilibrium()

# the cherab package has a convenience tool for viewing an equilibrium
# this function samples the various equilibrium attributes and renders them as images
print('Plotting equilibrium data...')
plot_equilibrium(equilibrium, detail=True)

# The core equilibrium properties are accessible as attributes on the the equilibrium object.
# For a full list of available attributes, see the object itself

# Psi and normalised psi are returned as 2D functions (see cherab.core.math.function.Function2D),
# they are cubically interpolated version of the raw gridded data. e.g. to obtain a value of psi
# normalised at the point (r=2.1m, z=0.2m), simply call the function:
psi_n = equilibrium.psi_normalised(2.1, 0.2)
print('Psi normalised at (2.1, 0.2)m: {}'.format(psi_n))

# The magnetic field is similarly accessible, it is returned as a 2D vector function. e.g. to
# obtain the components of the field at the point (r=2.1m, z=0.2m), simply call the function:
b = equilibrium.b_field(2.1, 0.2)
print('B-field at (2.1, 0.2)m: {}'.format(b))

# The inside_lcfs attribute returns a 2D function that identifies if a point lies inside the
# last closed flux surface. The function returns a value of 1.0 inside and a value of 0.0
# outside the LCFS.
print('Inside LCFS at (2.1, 0.2)m: {}'.format(equilibrium.inside_lcfs(2.1, 0.2)))
print('Inside LCFS at (2.1, 1.5)m: {}'.format(equilibrium.inside_lcfs(2.1, 1.5)))

# The equilibrium object includes methods for mapping functions onto the flux surfaces.
# These create 2D or 3D functions or e.g. temperature, density etc... according to the profile
# being mapped. The user can supply either a python function (with 1 argument - normalised psi),
# a Function1D object or a numerical array holding the normalised psi and function values.

# In this example we create fake 2D and 3D "temperature" profiles from an array of data. The
# array is interpolated with cubic interpolation and then mapped onto the normalised psi grid.
temperature_2d = equilibrium.map2d([[0, 0.5, 0.9, 1.0], [2500, 2000, 1000, 0]])
temperature_3d = equilibrium.map3d([[0, 0.5, 0.9, 1.0], [2500, 2000, 1000, 0]])

# display 2D temperature
print("Plotting array based 2d temperature...")

rmin, rmax = equilibrium.r_range
zmin, zmax = equilibrium.z_range
nr = round((rmax - rmin) / 0.025)
nz = round((zmax - zmin) / 0.025)
r, z, temperature_grid = sample2d(temperature_2d, (rmin, rmax, nr), (zmin, zmax, nz))

plt.figure()
plt.axes(aspect='equal')
plt.pcolormesh(r, z, temperature_grid.transpose(), shading='gouraud')
plt.autoscale(tight=True)
plt.colorbar()
plt.title('2D Temperature (array)')

# display 3D temperature
print("Plotting array based 3d temperature...")

rmin, rmax = equilibrium.r_range
zmin, zmax = equilibrium.z_range
nr = round((rmax - rmin) / 0.025)
nz = round((zmax - zmin) / 0.025)
temperature_slice = Slice3D(temperature_3d, axis='z', value=0.0)
x, y, temperature_grid = sample2d(temperature_slice, (-rmax, rmax, nr), (-rmax, rmax, nr))

plt.figure()
plt.axes(aspect='equal')
plt.pcolormesh(x, y, temperature_grid.transpose(), shading='gouraud')
plt.autoscale(tight=True)
plt.colorbar()
plt.title('3D Temperature (x-y slice, array)')

# In this example we interpolate the temperature data manually to produce a continuous function
# and then map that function around the flux surfaces to give the same result
te_psin = Interpolator1DArray([0, 0.5, 0.9, 1.0], [2500, 2000, 1000, 0], 'cubic', 'none', 0)

# map to produce 2D and 3D temperature profiles
temperature_2d = equilibrium.map2d(te_psin)
temperature_3d = equilibrium.map3d(te_psin)

# display 2D temperature
print("Plotting function based 2d temperature...")

rmin, rmax = equilibrium.r_range
zmin, zmax = equilibrium.z_range
nr = round((rmax - rmin) / 0.025)
nz = round((zmax - zmin) / 0.025)
r, z, temperature_grid = sample2d(temperature_2d, (rmin, rmax, nr), (zmin, zmax, nz))

plt.figure()
plt.axes(aspect='equal')
plt.pcolormesh(r, z, temperature_grid.transpose(), shading='gouraud')
plt.autoscale(tight=True)
plt.colorbar()
plt.title('2D Temperature (function)')

# display 3D temperature
print("Plotting function based 3d temperature...")

rmin, rmax = equilibrium.r_range
zmin, zmax = equilibrium.z_range
nr = round((rmax - rmin) / 0.01)
nz = round((zmax - zmin) / 0.01)
temperature_slice = Slice3D(temperature_3d, axis='z', value=0.0)
x, y, temperature_grid = sample2d(temperature_slice, (-rmax, rmax, nr), (-rmax, rmax, nr))

plt.figure()
plt.axes(aspect='equal')
plt.pcolormesh(x, y, temperature_grid.transpose(), shading='gouraud')
plt.autoscale(tight=True)
plt.colorbar()
plt.title('3D Temperature (x-y slice, function)')

plt.show()



# Copyright 2016-2022 Euratom
# Copyright 2016-2022 United Kingdom Atomic Energy Authority
# Copyright 2016-2022 Centro de Investigaciones Energéticas, Medioambientales y Tecnológicas
#
# Licensed under the EUPL, Version 1.1 or – as soon they will be approved by the
# European Commission - subsequent versions of the EUPL (the "Licence");
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at:
#
# https://joinup.ec.europa.eu/software/page/eupl5
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the Licence is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.
#
# See the Licence for the specific language governing permissions and limitations
# under the Licence.

from libc.math cimport M_PI

cdef:
    double RECIP_2_PI
    double RECIP_4_PI
    double DEGREES_TO_RADIANS
    double RADIANS_TO_DEGREES
    double ATOMIC_MASS
    double ELEMENTARY_CHARGE
    double SPEED_OF_LIGHT
    double PLANCK_CONSTANT
    double HC_EV_NM
    double ELECTRON_CLASSICAL_RADIUS
    double ELECTRON_REST_MASS
    double RYDBERG_CONSTANT_EV
    double VACUUM_PERMITTIVITY
    double BOHR_MAGNETON
